from poisson_approval.constants.basic_constants import APPROVAL, PLURALITY, ANTI_PLURALITY, \
    CANDIDATES, PAIRS_WITHOUT_INVERSIONS, BALLOTS_WITHOUT_INVERSIONS_SORTED_ALPHABETICAL


def allowed_ballots(voting_rule=APPROVAL):
    """Allowed ballots in a voting rule.

    Parameters
    ----------
    voting_rule : str
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``.

    Returns
    -------
    list
        The list of all allowed ballots.

    Examples
    --------
        >>> allowed_ballots(APPROVAL)
        ['a', 'ab', 'ac', 'b', 'bc', 'c']
        >>> allowed_ballots(PLURALITY)
        ['a', 'b', 'c']
        >>> allowed_ballots(ANTI_PLURALITY)
        ['ab', 'ac', 'bc']
    """
    if voting_rule == APPROVAL:
        return BALLOTS_WITHOUT_INVERSIONS_SORTED_ALPHABETICAL
    elif voting_rule == PLURALITY:
        return CANDIDATES
    elif voting_rule == ANTI_PLURALITY:
        return PAIRS_WITHOUT_INVERSIONS
    else:
        raise NotImplementedError


def ballot_one(ranking):
    """Ballot for the voter's preferred candidate.

    Parameters
    ----------
    ranking : str
        A ranking.

    Returns
    -------
    str
        The first candidate.

    Examples
    --------
        >>> ballot_one('abc')
        'a'
    """
    return ranking[0]


def ballot_two(ranking):
    """Ballot for the voter's second most-liked candidate.

    Parameters
    ----------
    ranking : str
        A ranking.

    Returns
    -------
    str
        The second candidate.

    Examples
    --------
        >>> ballot_two('abc')
        'b'
    """
    return ranking[1]


def ballot_one_two(ranking):
    """Ballot for the voter's two preferred candidates.

    Parameters
    ----------
    ranking : str
        A ranking.

    Returns
    -------
    str
        The ballot with the two first candidates.

    Examples
    --------
        >>> ballot_one_two('abc')
        'ab'
        >>> ballot_one_two('bac')
        'ab'
    """
    return sort_ballot(ranking[:2])


def ballot_one_three(ranking):
    """Ballot for the voter's first and third candidates.

    Parameters
    ----------
    ranking : str
        A ranking.

    Returns
    -------
    str
        The ballot with the first and third candidates.

    Examples
    --------
        >>> ballot_one_three('abc')
        'ac'
        >>> ballot_one_three('cba')
        'ac'
    """
    return sort_ballot(ranking[0] + ranking[2])


def ballot_low_u(ranking, voting_rule):
    """Ballot chosen by the voters who have a low utility for their middle candidate.

    Parameters
    ----------
    ranking : str
        A ranking.
    voting_rule : str
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``.

    Returns
    -------
    str
        The ballot chosen by the voters with this ranking and a low utility for their middle candidate, in case
        the response is utility-dependent.

    Examples
    --------
        >>> ballot_low_u('abc', APPROVAL)
        'a'
        >>> ballot_low_u('abc', PLURALITY)
        'a'
        >>> ballot_low_u('abc', ANTI_PLURALITY)
        'ac'
    """
    if voting_rule in {APPROVAL, PLURALITY}:
        return ballot_one(ranking)
    elif voting_rule == ANTI_PLURALITY:
        return ballot_one_three(ranking)
    else:
        raise NotImplementedError


def ballot_high_u(ranking, voting_rule):
    """Ballot chosen by the voters who have a high utility for their middle candidate.

    Parameters
    ----------
    ranking : str
        A ranking.
    voting_rule : str
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``.

    Returns
    -------
    str
        The ballot chosen by the voters with this ranking and a high utility for their middle candidate, in case
        the response is utility-dependent.

    Examples
    --------
        >>> ballot_high_u('abc', APPROVAL)
        'ab'
        >>> ballot_high_u('abc', PLURALITY)
        'b'
        >>> ballot_high_u('abc', ANTI_PLURALITY)
        'ab'
    """
    if voting_rule in {APPROVAL, ANTI_PLURALITY}:
        return ballot_one_two(ranking)
    elif voting_rule == PLURALITY:
        return ballot_two(ranking)
    else:
        raise NotImplementedError


def sort_ballot(ballot):
    """Put a ballot in alphabetical order.

    Parameters
    ----------
    ballot : str
        A ballot, e.g. ``'a'``, ``'ab'``, ``'ba'``, etc.

    Returns
    -------
    str
        The same ballot in alphabetical order.

    Examples
    --------
        >>> sort_ballot('a')
        'a'
        >>> sort_ballot('ab')
        'ab'
        >>> sort_ballot('ba')
        'ab'
    """
    return ''.join(sorted(ballot))



from poisson_approval import ProfileOrdinal


def is_not_condorcet(profile):
    """
    Whether a profile has no Condorcet winner.

    Parameters
    ----------
    profile : Profile
        A profile.

    Returns
    -------
    bool
        True if the profile has no Condorcet winner.

    Examples
    --------
        >>> profile = ProfileOrdinal({'abc': 0.3, 'bca': 0.4, 'cab': 0.3})
        >>> is_not_condorcet(profile)
        True

    This is just another syntax for:

        >>> profile.is_profile_condorcet == 0.0
        True

    This function is especially convenient when a test function is required, for example when using
    :func:`~poisson_approval.utils.Util.probability`.
    """
    return profile.is_profile_condorcet == 0.0



import numpy as np
from matplotlib import pyplot as plt
from poisson_approval.utils.UtilPlot import plt_cdf
from poisson_approval.random_factories.RandProfileHistogramUniform import RandProfileHistogramUniform
from poisson_approval.meta_analysis.monte_carlo_fictitious_play import monte_carlo_fictitious_play, \
    MCS_DECREASING_SCORES


def plot_distribution_scores(results, voting_rule=None, **kwargs):
    """
    Plot the distribution (CDF) of the scores for the winner, the challenger and the loser.

    Parameters
    ----------
    results : dict
        Results of :func:`~poisson_approval.meta_analysis.monte_carlo_fictitious_play.monte_carlo_fictitious_play`,
        with at least the setting
        :const:`~poisson_approval.meta_analysis.monte_carlo_fictitious_play.MCS_DECREASING_SCORES`.
    voting_rule : str
        The voting rule (or None if this parameter was not given to
        :func:`~poisson_approval.meta_analysis.monte_carlo_fictitious_play.monte_carlo_fictitious_play`).
    kwargs
        Other keyword arguments are passed to the function ``step`` of `matplotlib`.

    Examples
    --------
        >>> meta_results = monte_carlo_fictitious_play(
        ...     factory=RandProfileHistogramUniform(n_bins=1),
        ...     n_samples=1,
        ...     n_max_episodes=10,
        ...     monte_carlo_settings=[MCS_DECREASING_SCORES],
        ... )
        >>> plot_distribution_scores(meta_results)
    """
    if voting_rule is None:
        voting_rule = ''
    n_samples = results[voting_rule]['n_samples']
    weights = np.ones(n_samples) / n_samples
    # Plot the data
    fig, ax = plt.subplots(figsize=(8, 4))
    plt_cdf(results[voting_rule]['score_winner'], weights, n_samples, label='Winner', **kwargs)
    plt_cdf(results[voting_rule]['score_second'], weights, n_samples, label='Challenger', **kwargs)
    plt_cdf(results[voting_rule]['score_loser'], weights, n_samples, label='Loser', **kwargs)
    ax.grid(True)
    ax.legend()
    # ax.set_title('Score of the candidates')
    ax.set_xlabel('Score')
    ax.set_ylabel('Cumulative likelihood of occurrence')
    plt.xlim(0, 1)
    plt.xticks(np.arange(0, 1.1, 0.1))



import warnings
import itertools
from fractions import Fraction
from poisson_approval.constants.basic_constants import *
from poisson_approval.constants.EquilibriumStatus import EquilibriumStatus
from poisson_approval.iterables.IterableStrategyTwelve import IterableStrategyTwelve
from poisson_approval.profiles.ProfileCardinal import ProfileCardinal
from poisson_approval.strategies.StrategyTwelve import StrategyTwelve
from poisson_approval.tau_vector.TauVector import TauVector
from poisson_approval.utils.DictPrintingInOrder import DictPrintingInOrder
from poisson_approval.utils.DictPrintingInOrderIgnoringZeros import DictPrintingInOrderIgnoringZeros
from poisson_approval.utils.SetPrintingInOrder import SetPrintingInOrder
from poisson_approval.utils.Util import my_division
from poisson_approval.utils.UtilPreferences import sort_weak_order, is_hater
from poisson_approval.utils.UtilBallots import ballot_one, ballot_two, ballot_one_two, ballot_one_three, \
    ballot_low_u, ballot_high_u
from poisson_approval.utils.UtilCache import cached_property


# noinspection PyUnresolvedReferences
class ProfileTwelve(ProfileCardinal):
    """A profile of preference with twelve types.

    Parameters
    ----------
    d_type_share : dict
        E.g. ``{'ab_c': 0.4, 'c_ab': 0.6}``. ``d_type_share['ab_c']`` is the probability
        that a voter prefers candidate ``a``, then candidate ``b``, then candidate ``c``, with a utility for ``b``
        that is infinitely close to 1. In contrast, ``d_type_share['a_bc']`` is the probability that a voter prefers
        ``a`` then ``b`` then ``c``, with a utility for ``b`` that is infinitely close to 0.
    d_weak_order_share : dict
        E.g. ``{'a~b>c': 0.2, 'a>b~c': 0.1}``. ``d_weak_order_share['a~b>c']`` is the probability that a voter likes
        candidates ``a`` and ``b`` equally and prefer them to candidate ``c``.
    normalization_warning : bool
        Whether a warning should be issued if the input distribution is not normalized.
    ratio_sincere : Number
        The ratio of sincere voters, in the interval [0, 1]. This is used for :meth:`tau`.
    ratio_fanatic : Number
        The ratio of fanatic voters, in the interval [0, 1]. This is used for :meth:`tau`. The sum of `ratio_sincere`
        and `ratio_fanatic` must not exceed 1.
    voting_rule : str
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``.
    symbolic : bool
        Whether the computations are symbolic or numeric.

    Notes
    -----
    If the input distribution `d_type_share` is not normalized, the profile will be normalized anyway and a
    warning will be issued (unless `normalization_warning` is False).

    Examples
    --------
        >>> from fractions import Fraction
        >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
        ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
        >>> profile
        ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(3, 5), 'c_ab': Fraction(1, 5), \
'ca_b': Fraction(1, 10)})
        >>> print(profile)
        <ab_c: 1/10, b_ac: 3/5, c_ab: 1/5, ca_b: 1/10> (Condorcet winner: b)
        >>> profile.c_ab
        Fraction(1, 5)
        >>> profile.d_type_share['c_ab']  # Alternate syntax for profile.c_ab
        Fraction(1, 5)
        >>> profile.cab
        Fraction(3, 10)
        >>> profile.d_ranking_share['cab']  # Alternate syntax for profile.cab
        Fraction(3, 10)
        >>> profile.d_candidate_welfare
        {'a': Fraction(1, 5), 'b': Fraction(7, 10), 'c': Fraction(3, 10)}
        >>> profile.weighted_maj_graph
        array([[0, Fraction(-1, 5), Fraction(2, 5)],
               [Fraction(1, 5), 0, Fraction(2, 5)],
               [Fraction(-2, 5), Fraction(-2, 5), 0]], dtype=object)
        >>> profile.condorcet_winners
        Winners({'b'})
        >>> profile.is_profile_condorcet
        1.0
        >>> profile.has_majority_favorite  # Is one candidate 'top' in a majority of ballots?
        True
        >>> profile.has_majority_ranking  # Does one ranking represent a majority of ballots?
        True
        >>> profile.is_single_peaked  # Is the profile single-peaked?
        True
        >>> profile.support_in_rankings
        {'abc', 'bac', 'cab'}
        >>> profile.is_generic_in_rankings  # Are all rankings there?
        False
        >>> profile.analyzed_strategies_pure
        Equilibria:
        <abc: a, bac: b, cab: ac> ==> b (FF)
        <abc: a, bac: ab, cab: c> ==> a (D)
        <abc: ab, bac: b, cab: utility-dependent> ==> b (FF)
        <BLANKLINE>
        Non-equilibria:
        <abc: a, bac: b, cab: c> ==> b (FF)
        <abc: a, bac: b, cab: utility-dependent> ==> b (FF)
        <abc: a, bac: ab, cab: ac> ==> a (D)
        <abc: a, bac: ab, cab: utility-dependent> ==> a (D)
        <abc: ab, bac: b, cab: c> ==> b (FF)
        <abc: ab, bac: b, cab: ac> ==> b (FF)
        <abc: ab, bac: ab, cab: c> ==> a, b (FF)
        <abc: ab, bac: ab, cab: ac> ==> a (D)
        <abc: ab, bac: ab, cab: utility-dependent> ==> a (D)
        >>> print(profile.analyzed_strategies_pure.winners_at_equilibrium)
        a, b

    In the following example, one third of the voters are sincere:

        >>> from fractions import Fraction
        >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
        ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)},
        ...                         ratio_sincere=Fraction(1, 3))
        >>> tau_sincere = profile.tau_sincere
        >>> print(tau_sincere)
        <ab: 1/10, ac: 1/10, b: 3/5, c: 1/5> ==> b
        >>> strategy = StrategyTwelve({'abc': 'a', 'bac': 'b', 'cab': 'utility-dependent'})
        >>> tau_strategic = profile.tau_strategic(strategy)
        >>> print(tau_strategic)
        <a: 1/10, ac: 1/10, b: 3/5, c: 1/5> ==> b
        >>> tau = profile.tau(strategy)
        >>> print(tau)
        <a: 1/15, ab: 1/30, ac: 1/10, b: 3/5, c: 1/5> ==> b

    The profile can include weak orders:

        >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10)},
        ...                         d_weak_order_share={'a~b>c': Fraction(3, 10)})
        >>> profile
        ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(3, 5)}, d_weak_order_share={'a~b>c': Fraction(3, 10)})
        >>> print(profile)
        <ab_c: 1/10, b_ac: 3/5, a~b>c: 3/10> (Condorcet winner: b)
    """

    def __init__(self, d_type_share, d_weak_order_share=None, normalization_warning=True,
                 ratio_sincere=0, ratio_fanatic=0, voting_rule=APPROVAL, symbolic=False):
        super().__init__(ratio_sincere=ratio_sincere, ratio_fanatic=ratio_fanatic, voting_rule=voting_rule,
                         symbolic=symbolic)
        if d_weak_order_share is None:
            d_weak_order_share = dict()
        # Populate the dictionaries of types and weak orders
        self.d_type_share = DictPrintingInOrderIgnoringZeros({t: 0 for t in TWELVE_TYPES})
        self._d_weak_order_share = DictPrintingInOrderIgnoringZeros({
            weak_order: 0 for weak_order in WEAK_ORDERS_WITHOUT_INVERSIONS})
        for t, share in itertools.chain(d_type_share.items(), d_weak_order_share.items()):
            try:
                self.d_type_share[t] += share
            except KeyError:
                self._d_weak_order_share[sort_weak_order(t)] += share
        # Normalize if necessary
        total = sum(self.d_type_share.values()) + sum(self._d_weak_order_share.values())
        if not self.ce.look_equal(total, 1):
            if normalization_warning:
                warnings.warn(NORMALIZATION_WARNING)
            for t in self.d_type_share.keys():
                self.d_type_share[t] = my_division(self.d_type_share[t], total)
            for weak_order in self._d_weak_order_share.keys():
                self._d_weak_order_share[weak_order] = my_division(self._d_weak_order_share[weak_order], total)

    @cached_property
    def d_weak_order_share(self):
        return self._d_weak_order_share

    def have_ranking_with_utility_above_u(self, ranking, u):
        """Share of voters who have a given ranking and a utility for their middle candidate that is strictly above a
        given value.

        Cf. :meth:`ProfileCardinal.have_ranking_with_utility_above_u`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> profile.have_ranking_with_utility_above_u(ranking='cab', u=.5)
            Fraction(1, 10)
            >>> profile.have_ranking_with_utility_above_u(ranking='cab', u=0)
            Fraction(3, 10)
            >>> profile.have_ranking_with_utility_above_u(ranking='cab', u=1)
            0
        """
        if u == 1:
            return 0
        high_u = self.d_type_share[ranking[:2] + '_' + ranking[2:]]  # E.g. ab_c
        if u == 0:
            low_u = self.d_type_share[ranking[:1] + '_' + ranking[1:]]  # E.g. a_bc
            return high_u + low_u
        return high_u

    def have_ranking_with_utility_u(self, ranking, u):
        """Share of voters who have a given ranking and a utility for their middle candidate that is equal to a given
        value.

        Cf. :meth:`ProfileCardinal.have_ranking_with_utility_u`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> profile.have_ranking_with_utility_u(ranking='cab', u=.5)
            0
        """
        return 0

    def have_ranking_with_utility_below_u(self, ranking, u):
        """Share of voters who have a given ranking and a utility for their middle candidate that is strictly below a
        given value.

        Cf. :meth:`ProfileCardinal.have_ranking_with_utility_below_u`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> profile.have_ranking_with_utility_below_u(ranking='cab', u=.5)
            Fraction(1, 5)
            >>> profile.have_ranking_with_utility_below_u(ranking='cab', u=1)
            Fraction(3, 10)
            >>> profile.have_ranking_with_utility_below_u(ranking='cab', u=0)
            0
        """
        if u == 0:
            return 0
        low_u = self.d_type_share[ranking[:1] + '_' + ranking[1:]]   # E.g. a_bc
        if u == 1:
            high_u = self.d_type_share[ranking[:2] + '_' + ranking[2:]]  # E.g. ab_c
            return high_u + low_u
        return low_u

    def __repr__(self):
        """
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)},
            ...                         ratio_sincere=Fraction(1, 10), ratio_fanatic=Fraction(1, 5))
            >>> profile
            ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(3, 5), 'c_ab': Fraction(1, 5), \
'ca_b': Fraction(1, 10)}, ratio_sincere=Fraction(1, 10), ratio_fanatic=Fraction(1, 5))
        """
        arguments = repr(self.d_type_share)
        if self.contains_weak_orders:
            arguments += ', d_weak_order_share=%r' % self.d_weak_order_share
        if self.ratio_sincere > 0:
            arguments += ', ratio_sincere=%r' % self.ratio_sincere
        if self.ratio_fanatic > 0:
            arguments += ', ratio_fanatic=%r' % self.ratio_fanatic
        if self.voting_rule != APPROVAL:
            arguments += ', voting_rule=%r' % self.voting_rule
        return 'ProfileTwelve(%s)' % arguments

    def __str__(self):
        """
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)},
            ...                         ratio_sincere=Fraction(1, 10), ratio_fanatic=Fraction(1, 5))
            >>> print(profile)
            <ab_c: 1/10, b_ac: 3/5, c_ab: 1/5, ca_b: 1/10> (Condorcet winner: b) (ratio_sincere: 1/10) \
(ratio_fanatic: 1/5)
        """
        contents = []
        if self.contains_rankings:
            contents.append(str(self.d_type_share)[1:-1])
        if self.contains_weak_orders:
            contents.append(str(self.d_weak_order_share)[1:-1])
        result = '<' + ', '.join(contents) + '>'
        if self.is_profile_condorcet:
            result += ' (Condorcet winner: %s)' % self.condorcet_winners
        if self.ratio_sincere > 0:
            result += ' (ratio_sincere: %s)' % self.ratio_sincere
        if self.ratio_fanatic > 0:
            result += ' (ratio_fanatic: %s)' % self.ratio_fanatic
        if self.voting_rule != APPROVAL:
            result += ' (%s)' % self.voting_rule
        return result

    def _repr_pretty_(self, p, cycle):  # pragma: no cover - Only for notebooks
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')

    def __eq__(self, other):
        """Equality test.

        Parameters
        ----------
        other : object

        Returns
        -------
        bool
            True iff this profile is equal to `other`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> profile == ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                           'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            True
        """
        return (isinstance(other, ProfileTwelve)
                and self.d_type_share == other.d_type_share
                and self.d_weak_order_share == other.d_weak_order_share
                and self.ratio_sincere == other.ratio_sincere
                and self.ratio_fanatic == other.ratio_fanatic
                and self.voting_rule == other.voting_rule)

    # Standardized version of the profile (makes it unique, up to permutations)

    @cached_property
    def standardized_version(self):
        """ProfileTwelve : Standardized version of the profile (makes it unique, up to permutations of the candidates).

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> print(profile.standardized_version)
            <a_bc: 3/5, ba_c: 1/10, c_ba: 1/5, cb_a: 1/10> (Condorcet winner: a)
            >>> profile.is_standardized
            False
        """
        def translate(s, permute):
            return s.replace('a', permute[0]).replace('b', permute[1]).replace('c', permute[2])

        best_d = {}
        best_signature = []
        for perm in XYZ_PERMUTATIONS:
            d_test = {translate(t, perm): share for t, share in self.d_type_share.items()}
            d_test.update({sort_weak_order(translate(weak_order, perm)): share
                           for weak_order, share in self.d_weak_order_share.items()})
            signature_test = [d_test[t] for t in XYZ_TWELVE_TYPES]
            signature_test += [d_test[weak_order] for weak_order in XYZ_WEAK_ORDERS_WITHOUT_INVERSIONS]
            if signature_test > best_signature:
                best_signature = signature_test
                best_d = d_test
        return ProfileTwelve({t: best_d[xyz_t] for t, xyz_t in zip(TWELVE_TYPES, XYZ_TWELVE_TYPES)},
                             {weak_order: best_d[xyz_weak_order] for weak_order, xyz_weak_order in zip(
                                 WEAK_ORDERS_WITHOUT_INVERSIONS, XYZ_WEAK_ORDERS_WITHOUT_INVERSIONS)},
                             ratio_sincere=self.ratio_sincere, ratio_fanatic=self.ratio_fanatic,
                             voting_rule=self.voting_rule)

    @cached_property
    def d_candidate_welfare(self):
        d = DictPrintingInOrder({candidate: 0 for candidate in CANDIDATES})
        for ranking in RANKINGS:
            share_i_jk = self.d_type_share[ranking[:1] + '_' + ranking[1:]]  # E.g. a_bc
            share_ij_k = self.d_type_share[ranking[:2] + '_' + ranking[2:]]  # E.g. ab_c
            d[ranking[0]] += share_i_jk + share_ij_k
            d[ranking[1]] += share_ij_k
        for weak_order in self.support_in_weak_orders:
            share = self.d_weak_order_share[weak_order]
            d[weak_order[0]] += share
            if is_hater(weak_order):
                d[weak_order[2]] += share
        return d

    @cached_property
    def has_majority_type(self):
        """bool : Whether there is a majority type (a type shared by strictly more than half of the voters).

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> profile.has_majority_type
            True

        This does **not** include weak orders:

            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10)}, d_weak_order_share={'b>a~c': Fraction(9, 10)})
            >>> profile.has_majority_type
            False
        """
        return max(self.d_type_share.values()) > Fraction(1, 2)

    # Has full support
    @cached_property
    def support_in_types(self):
        """:class:`SetPrintingInOrder` of str : Support of the profile (in terms of types).

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> profile.support_in_types
            {'ab_c', 'b_ac', 'c_ab', 'ca_b'}
        """
        return SetPrintingInOrder({t for t, share in self.d_type_share.items() if share > 0})

    @cached_property
    def is_generic_in_types(self):
        """bool : Whether the profile is generic in types (contains all types).

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> profile.is_generic_in_types
            False
        """
        return 0 not in self.d_type_share.values()

    # Tau and strategy-related stuff

    def tau_strategic(self, strategy):
        """Tau-vector associated to a strategy.

        Parameters
        ----------
        strategy : StrategyTwelve
            A strategy that specifies at least all the rankings that are present in the profile.

        Returns
        -------
        TauVector
            Tau-vector associated to this profile and strategy `strategy`.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> strategy = StrategyTwelve({'abc': 'ab', 'bac': 'b', 'cab': 'utility-dependent'})
            >>> tau_strategic = profile.tau_strategic(strategy)
            >>> print(tau_strategic)
            <ab: 1/10, ac: 1/10, b: 3/5, c: 1/5> ==> b
        """
        assert self.voting_rule == strategy.voting_rule
        t = self.d_ballot_share_weak_voters_strategic(strategy)
        for ranking, ballot in strategy.d_ranking_ballot.items():
            if self.d_ranking_share[ranking] == 0:
                continue
            # For a ranking abc, ballot can be real ballots (e.g. 'a', 'ab'), '' or 'utility-dependent'.
            if ballot == UTILITY_DEPENDENT:
                t[ballot_low_u(ranking, self.voting_rule)] += self.have_ranking_with_utility_below_u(ranking, u=.5)
                t[ballot_high_u(ranking, self.voting_rule)] += self.have_ranking_with_utility_above_u(ranking, u=.5)
            else:
                t[ballot] += self.d_ranking_share[ranking]
        return TauVector(t, voting_rule=self.voting_rule, symbolic=self.symbolic)

    def share_sincere_among_strategic_voters(self, strategy):
        """Share of strategic voters that happen to cast a sincere ballot.

        Parameters
        ----------
        strategy : StrategyTwelve
            A strategy that specifies at least all the rankings that are present in the profile.

        Returns
        -------
        Number
            The ratio of sincere voters among strategic voters.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> strategy = StrategyTwelve({'abc': 'ab', 'bac': 'b', 'cab': 'utility-dependent'})
            >>> profile.share_sincere_among_strategic_voters(strategy)
            1
        """
        assert self.voting_rule == strategy.voting_rule
        # Weak orders: these voters are always "sincere", even in Plurality or Anti-Plurality. For example, in
        # Plurality, a voter a~b>c will either vote for a or b, and this is considered "sincere".
        share_sincere = sum(self.d_weak_order_share.values())
        # Rankings
        for ranking, ballot in strategy.d_ranking_ballot.items():
            if self.d_ranking_share[ranking] == 0:
                continue
            # For a ranking abc, ballot can be real ballots (e.g. 'a', 'ab'), '' or 'utility-dependent'.
            if self.voting_rule == APPROVAL:
                if ballot == ballot_one(ranking):
                    share_sincere += self.have_ranking_with_utility_below_u(ranking, u=.5)
                elif ballot == ballot_one_two(ranking):
                    share_sincere += self.have_ranking_with_utility_above_u(ranking, u=.5)
                else:
                    share_sincere += self.d_ranking_share[ranking]
            elif self.voting_rule == PLURALITY:
                if ballot == ballot_one(ranking):
                    share_sincere += self.d_ranking_share[ranking]
                elif ballot == UTILITY_DEPENDENT:
                    share_sincere += self.have_ranking_with_utility_below_u(ranking, u=.5)
            elif self.voting_rule == ANTI_PLURALITY:
                if ballot == ballot_one_two(ranking):
                    share_sincere += self.d_ranking_share[ranking]
                elif ballot == UTILITY_DEPENDENT:
                    share_sincere += self.have_ranking_with_utility_above_u(ranking, u=.5)
            else:
                raise NotImplementedError
        return self.ce.simplify(share_sincere)

    def is_equilibrium(self, strategy):
        """Whether a strategy is an equilibrium.

        Parameters
        ----------
        strategy : StrategyTwelve
            A strategy that specifies at least all the rankings that are present in the profile.

        Returns
        -------
        EquilibriumStatus
            Whether `strategy` is an equilibrium in this profile.

        Examples
        --------
            >>> from fractions import Fraction
            >>> profile = ProfileTwelve({'ab_c': Fraction(1, 10), 'b_ac': Fraction(6, 10),
            ...                          'c_ab': Fraction(2, 10), 'ca_b': Fraction(1, 10)})
            >>> strategy = StrategyTwelve({'abc': 'ab', 'bac': 'b', 'cab': 'utility-dependent'})
            >>> profile.is_equilibrium(strategy)
            EquilibriumStatus.EQUILIBRIUM
        """
        d_ranking_best_response = self.tau(strategy).d_ranking_best_response
        status = EquilibriumStatus.EQUILIBRIUM
        for ranking, share in self.d_ranking_share.items():
            if share == 0:
                continue
            best_response = d_ranking_best_response[ranking]
            if best_response.ballot == INCONCLUSIVE:  # pragma: no cover - Should never happen
                status = min(status, EquilibriumStatus.INCONCLUSIVE)
            else:
                type_1 = ranking[:1] + '_' + ranking[1:]  # E.g. a_bc
                type_12 = ranking[:2] + '_' + ranking[2:]  # E.g. ab_c
                if best_response.ballot == UTILITY_DEPENDENT:
                    if self.voting_rule == APPROVAL:
                        best_ballot_1, best_ballot_12 = ballot_one(ranking), ballot_one_two(ranking)
                    elif self.voting_rule == PLURALITY:
                        best_ballot_1, best_ballot_12 = ballot_one(ranking), ballot_two(ranking)
                    elif self.voting_rule == ANTI_PLURALITY:
                        best_ballot_1, best_ballot_12 = ballot_one_three(ranking), ballot_one_two(ranking)
                    else:
                        raise NotImplementedError
                else:
                    best_ballot_1 = best_response.ballot
                    best_ballot_12 = best_response.ballot
                if self.d_type_share[type_1] > 0 and getattr(strategy, type_1) != best_ballot_1:
                    return EquilibriumStatus.NOT_EQUILIBRIUM
                if self.d_type_share[type_12] > 0 and getattr(strategy, type_12) != best_ballot_12:
                    return EquilibriumStatus.NOT_EQUILIBRIUM
        return status

    @property
    def strategies_pure(self):
        """Iterator: pure strategies of the profile.

        Yields
        ------
        :class:`StrategyTwelve`
            All possible pure strategies of the profile.
        """
        return IterableStrategyTwelve(profile=self)

    @property
    def strategies_group(self):
        raise NotImplementedError

    @classmethod
    def order_and_label(cls, t):
        r"""Order and label of a discrete type.

        Cf. :meth:`Profile.order_and_label`.

        Examples
        --------
            >>> ProfileTwelve.order_and_label('ab_c')
            ('abc', '$r(ab\\_c)$')
            >>> ProfileTwelve.order_and_label('a~b>c')
            ('a~b>c', '$r(a\\sim b>c)$')
        """
        if len(t) == 4:
            return t.replace('_', ''), ('$r(%s)$' % t).replace('_', '\\_')
        else:
            return cls.order_and_label_weak(t)


def make_property_type_share(t, doc):
    def _f(self):
        return self.d_type_share[t]
    _f.__doc__ = doc
    return property(_f)


for my_t in TWELVE_TYPES:
    setattr(ProfileTwelve, my_t, make_property_type_share(my_t, 'Number : Share of voters with this type.'))



from poisson_approval.constants.basic_constants import *
from poisson_approval.strategies.StrategyTwelve import StrategyTwelve
from poisson_approval.utils.DictPrintingInOrderIgnoringZeros import DictPrintingInOrderIgnoringZeros
from poisson_approval.utils.DictPrintingInOrderIgnoringNone import DictPrintingInOrderIgnoringNone
from poisson_approval.utils.UtilBallots import ballot_one, ballot_two, ballot_one_two, ballot_one_three
from poisson_approval.utils.UtilCache import cached_property


# noinspection PyUnresolvedReferences
class StrategyThreshold(StrategyTwelve):
    """A threshold strategy (for a cardinal profile).

    For each ranking, there is a ``threshold`` and a ``ratio_optimistic``. E.g. assume that for ranking ``abc``,
    the threshold is 0.4 and the ratio is 0.2. It means that:

    * A voter ``abc`` votes for `ab` if their utility for `b` is strictly greater than 0.4,
    * A voter ``abc`` votes for `a` if their utility for `b` is strictly lower than 0.4,
    * Voters ``abc`` whose utility for `b` is equal to 0.4 are split: a ratio 0.2 are optimistic, they vote only for
      `a` (they behave as if the pivot `ab` was very likely); and a ratio 0.8 are pessimistic, they vote for `ab`
      (they behave as if the pivot `bc` was very likely).

    For a given ranking, the threshold and/or the ratio may be None, which means that they are not specified for
    this strategy.

    Parameters
    ----------
    d : dict
        Cf. examples below for the different types of input syntax.
    d_weak_order_ballot : dict
        Key: weak order. Value: strategy. A strategy can be a valid ballot, ``SPLIT`` or ``''`` if the behavior
        of these voters is not specified in the strategy. This is useful in two cases only: for "haters"
        (e.g. ``'a~b>c'``) in Plurality, and for "lovers" (e.g. ``'a>b~c'``) in Anti-Plurality.
        In all other cases, voters with a weak order have a dominant strategy and they automatically use it.
        About ``'SPLIT'``: for example, in Plurality, ``'a~b>c': SPLIT`` means that half of voters with weak order
        `abc` cast a ballot for `a`, and the other half for `b`.
    ratio_optimistic : Number
        Cf. examples below for the different types of input syntax.
    profile : Profile, optional
        The "context" in which the strategy is used.
    voting_rule : str
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``. Default: the same
        voting rule as `profile` if a profile is specified, ``APPROVAL`` otherwise.

    Examples
    --------
    The two following examples illustrate different ways to define the same profile. The first possible type of input
    syntax is a dict that maps a ranking to a tuple (``threshold``, ``ratio_optimistic``). It corresponds exactly to
    the attribute :attr:`d_ranking_t_threshold_ratio_optimistic`:

        >>> strategy = StrategyThreshold({'abc': (0.4, 0.2), 'bac': (0.51, 0.2), 'cab': (1, 0.2)})
        >>> print(strategy)
        <abc: utility-dependent (0.4, 0.2), bac: utility-dependent (0.51, 0.2), cab: c>

    The second  possible type of input syntax is a dict that maps a ranking to a threshold. All rankings have the same
    ratio, given by the parameter `ratio_optimistic`:

        >>> strategy = StrategyThreshold({'abc': 0.4, 'bac': 0.51, 'cab': 1}, ratio_optimistic=0.2)
        >>> print(strategy)
        <abc: utility-dependent (0.4, 0.2), bac: utility-dependent (0.51, 0.2), cab: c>

    Some operations on the strategy:

        >>> strategy
        StrategyThreshold({'abc': (0.4, 0.2), 'bac': (0.51, 0.2), 'cab': (1, 0.2)})
        >>> strategy.abc
        'utility-dependent'
        >>> strategy.a_bc
        'a'
        >>> strategy.ab_c
        'ab'
        >>> strategy.d_ranking_threshold['abc']
        0.4

    It is possible not to specify the ratios of optimistic voters:

        >>> strategy = StrategyThreshold({'abc': 0.4, 'bac': 0.51})
        >>> strategy
        StrategyThreshold({'abc': 0.4, 'bac': 0.51})
        >>> print(strategy)
        <abc: utility-dependent (0.4), bac: utility-dependent (0.51)>

    If this strategy is applied to a profile where a positive share of voters have ranking `abc` and a utility 0.2
    for their middle candidate `b`, it will raise an error because the ratio of optimistic voters is needed in that
    case.
    """

    def __init__(self, d, d_weak_order_ballot=None, ratio_optimistic=None, profile=None, voting_rule=None):
        voting_rule = self._get_voting_rule_(profile, voting_rule)
        # Prepare the dictionaries of thresholds and ratios
        self.d_ranking_threshold = DictPrintingInOrderIgnoringNone({ranking: None for ranking in RANKINGS})
        self.d_ranking_ratio_optimistic = DictPrintingInOrderIgnoringNone({ranking: None for ranking in RANKINGS})
        for ranking, value in d.items():
            if isinstance(value, tuple):
                self.d_ranking_threshold[ranking], self.d_ranking_ratio_optimistic[ranking] = value
            else:
                self.d_ranking_threshold[ranking] = value
                self.d_ranking_ratio_optimistic[ranking] = ratio_optimistic
        # Prepare the dictionary of ballots
        d_ranking_ballot = DictPrintingInOrderIgnoringZeros()
        for ranking, threshold in self.d_ranking_threshold.items():
            if threshold is None:
                d_ranking_ballot[ranking] = ''
            elif threshold == 1:
                if voting_rule in {APPROVAL, PLURALITY}:
                    d_ranking_ballot[ranking] = ballot_one(ranking)
                elif voting_rule == ANTI_PLURALITY:
                    d_ranking_ballot[ranking] = ballot_one_three(ranking)
                else:
                    raise NotImplementedError
            elif threshold == 0:
                if voting_rule in {APPROVAL, ANTI_PLURALITY}:
                    d_ranking_ballot[ranking] = ballot_one_two(ranking)
                elif voting_rule == PLURALITY:
                    d_ranking_ballot[ranking] = ballot_two(ranking)
                else:
                    raise NotImplementedError
            else:
                d_ranking_ballot[ranking] = UTILITY_DEPENDENT
        # Call parent class
        super().__init__(d_ranking_ballot=d_ranking_ballot, d_weak_order_ballot=d_weak_order_ballot,
                         profile=profile, voting_rule=voting_rule)

    @cached_property
    def d_ranking_t_threshold_ratio_optimistic(self):
        """dict : Thresholds and ratios of optimistic voters.

        Key : ranking, e.g. ``'abc'``. Value: tuple (``threshold``, ``ratio_optimistic``).
        """
        return DictPrintingInOrderIgnoringNone({
            ranking: (self.d_ranking_threshold[ranking], self.d_ranking_ratio_optimistic[ranking])
            for ranking in RANKINGS
        })

    def __eq__(self, other):
        """Equality test.

        Parameters
        ----------
        other : object

        Returns
        -------
        bool
            True if this strategy is equal to `other`.

        Examples
        --------
            >>> strategy = StrategyThreshold({'abc': (0.4, 0.2), 'bac': (0.51, 0.2), 'cab': (1, 0.2)})
            >>> strategy == StrategyThreshold({'abc': (0.4, 0.2), 'bac': (0.51, 0.2), 'cab': (1, 0.2)})
            True
        """
        return (isinstance(other, StrategyThreshold)
                and self.d_ranking_threshold == other.d_ranking_threshold
                and self.d_ranking_ratio_optimistic == other.d_ranking_ratio_optimistic
                and self.d_weak_order_ballot == other.d_weak_order_ballot
                and self.voting_rule == other.voting_rule)

    def __repr__(self):
        """
        Examples
        --------
            >>> strategy = StrategyThreshold({'abc': (0.4, 0.2)}, d_weak_order_ballot={'a~b>c': 'a'},
            ...                              voting_rule=PLURALITY)
            >>> repr(strategy)
            "StrategyThreshold({'abc': (0.4, 0.2)}, d_weak_order_ballot={'a~b>c': 'a'}, voting_rule='Plurality')"
        """
        d = DictPrintingInOrderIgnoringNone({
            k: t[0] if t[1] is None else t
            for k, t in self.d_ranking_t_threshold_ratio_optimistic.items()})
        arguments = repr(d)
        arguments_weak_orders = repr(self.d_weak_order_ballot)
        if len(arguments_weak_orders) > 2:
            arguments += ', d_weak_order_ballot=' + arguments_weak_orders
        if self.voting_rule != APPROVAL:
            arguments += ', voting_rule=%r' % self.voting_rule
        return 'StrategyThreshold(%s)' % arguments

    def __str__(self):
        """
        Examples
        --------
            >>> strategy = StrategyThreshold({'abc': (0.4, 0.2)}, d_weak_order_ballot={'a~b>c': 'a'},
            ...                              voting_rule=PLURALITY)
            >>> str(strategy)
            '<abc: utility-dependent (0.4, 0.2), a~b>c: a> (Plurality)'
        """
        def t_threshold_ratio_to_string(t):
            return str(t[0]) if t[1] is None else '%s, %s' % t

        arguments = ', '.join([
            '%s: %s' % (ranking, self.d_ranking_ballot[ranking])
            + (' (%s)' % t_threshold_ratio_to_string(self.d_ranking_t_threshold_ratio_optimistic[ranking])
               if self.d_ranking_ballot[ranking] == UTILITY_DEPENDENT else '')
            for ranking in sorted(self.d_ranking_ballot) if self.d_ranking_ballot[ranking]
        ])
        arguments_weak_orders = str(self.d_weak_order_ballot)[1:-1]
        if len(arguments_weak_orders) > 0:
            arguments += ', ' + arguments_weak_orders
        result = '<%s>' % arguments
        if self.profile is not None:
            result += ' ==> ' + str(self.winners)
        if self.voting_rule != APPROVAL:
            result += ' (%s)' % self.voting_rule
        return result

    def _repr_pretty_(self, p, cycle):  # pragma: no cover - Only for notebooks
        # https://stackoverflow.com/questions/41453624/tell-ipython-to-use-an-objects-str-instead-of-repr-for-output
        p.text(str(self) if not cycle else '...')



from poisson_approval import monte_carlo_fictitious_play, RandProfileHistogramUniform, \
    MCS_PROFILE, MCS_TAU_INIT, MCS_N_EPISODES, MCS_CANDIDATE_WINNING_FREQUENCY, MCS_CONVERGES, MCS_FREQUENCY_CW_WINS, \
    MCS_WELFARE_LOSSES, MCS_UTILITY_THRESHOLDS, MCS_BALLOT_STATISTICS, MCS_DECREASING_SCORES, VOTING_RULES, \
    one_over_t


def test_no_mcs():
    """
        >>> meta_results = monte_carlo_fictitious_play(
        ...     factory=RandProfileHistogramUniform(n_bins=1),
        ...     n_samples=1,
        ...     n_max_episodes=10,
        ... )
        >>> meta_results
        {'': {'n_samples': 1}}
    """
    pass


def test_with_save_file():
    """
        >>> import pickle
        >>> file_save = 'zzz_test_monte_carlo_with_save_file.sav'
        >>> _ = monte_carlo_fictitious_play(
        ...     factory=RandProfileHistogramUniform(n_bins=1),
        ...     n_samples=1,
        ...     n_max_episodes=10,
        ...     file_save=file_save,
        ... )
        >>> with open(file_save, "rb") as f:
        ...     meta_results = pickle.load(f)
        >>> meta_results
        {'': {'n_samples': 1}}
    """
    pass


def test_monte_carlo_iterated_voting():
    """
        >>> meta_results = monte_carlo_fictitious_play(
        ...     factory=RandProfileHistogramUniform(n_bins=1),
        ...     n_samples=1,
        ...     n_max_episodes=10,
        ...     voting_rules=VOTING_RULES,
        ...     monte_carlo_settings=[
        ...         MCS_PROFILE,
        ...         MCS_TAU_INIT,
        ...         MCS_N_EPISODES,
        ...         MCS_CANDIDATE_WINNING_FREQUENCY,
        ...         MCS_CONVERGES,
        ...         MCS_FREQUENCY_CW_WINS,
        ...         MCS_WELFARE_LOSSES,
        ...         MCS_UTILITY_THRESHOLDS,
        ...         MCS_BALLOT_STATISTICS,
        ...         MCS_DECREASING_SCORES,
        ...     ],
        ...     meth='iterated_voting',
        ...     perception_update_ratio=1,
        ...     ballot_update_ratio=1,
        ...     statistics_update_ratio=one_over_t,
        ... )
    """
    pass



import numpy as np
from poisson_approval.utils.DictPrintingInOrder import DictPrintingInOrder
from poisson_approval.utils.Util import iterate_simplex_grid


class IterableSimplexGrid:
    """Iterate over some objects defined by shares on a discrete simplex.

    Parameters
    ----------
    cls : class
        The class of object we want to create. It must accept as parameter a dictionary of the form ``key: share``,
        where share is a number.
    denominator : int or iterable
        The grain(s) of the grid.
    keys : iterable
        These keys will have a variable share.
    d_key_fixed_share : dict
        A dictionary. For each entry ``key: fixed_share``, this key will have at least this fixed share. The total
        must be lower or equal to 1.
    test : callable
        A function ``cls -> bool``.
    kwargs
        Additional parameters are passed to `cls` when creating the object.

    Examples
    --------
    Basic usage:

        >>> from fractions import Fraction
        >>> for d in IterableSimplexGrid(cls=DictPrintingInOrder, denominator=3, keys=['a', 'b']):
        ...     print(d)
        {a: 1, b: 0}
        {a: 2/3, b: 1/3}
        {a: 1/3, b: 2/3}
        {a: 0, b: 1}

    It is possible to specify an iterable of denominators:

        >>> from fractions import Fraction
        >>> for d in IterableSimplexGrid(cls=DictPrintingInOrder, denominator=range(2, 4), keys=['a', 'b']):
        ...     print(d)
        {a: 1, b: 0}
        {a: 1/2, b: 1/2}
        {a: 0, b: 1}
        {a: 1, b: 0}
        {a: 2/3, b: 1/3}
        {a: 1/3, b: 2/3}
        {a: 0, b: 1}

    If `d_key_fixed_share` is given, then these shares are fixed, and the remaining share is split between `keys`:

        >>> for d in IterableSimplexGrid(cls=DictPrintingInOrder, denominator=3, keys=['a', 'b'],
        ...                              d_key_fixed_share={'c': Fraction(1, 2)}):
        ...     print(d)
        {a: 1/2, b: 0, c: 1/2}
        {a: 1/3, b: 1/6, c: 1/2}
        {a: 1/6, b: 1/3, c: 1/2}
        {a: 0, b: 1/2, c: 1/2}

    The keys in `d_fixed_share` may overlap with `keys`:

        >>> for d in IterableSimplexGrid(cls=DictPrintingInOrder, denominator=3, keys=['a', 'b'],
        ...                              d_key_fixed_share={'b': Fraction(1, 2)}):
        ...     print(d)
        {a: 1/2, b: 1/2}
        {a: 1/3, b: 2/3}
        {a: 1/6, b: 5/6}
        {a: 0, b: 1}

    It is possible to add a condition with the parameter `test`:

        >>> def test_small_euclidean_norm(d):
        ...     return sum(x**2 for x in d.values()) < 0.6
        >>> for d in IterableSimplexGrid(cls=DictPrintingInOrder, denominator=11, keys=['a', 'b'],
        ...                              test=test_small_euclidean_norm):
        ...     print(d)
        {a: 7/11, b: 4/11}
        {a: 6/11, b: 5/11}
        {a: 5/11, b: 6/11}
        {a: 4/11, b: 7/11}
    """

    def __init__(self, cls, denominator, keys, d_key_fixed_share=None, test=None, **kwargs):
        # Default parameters
        if d_key_fixed_share is None:
            d_key_fixed_share = dict()
        # Parameters
        self.cls = cls
        self.denominator = denominator
        self.keys = keys
        self.d_key_fixed_share = d_key_fixed_share
        self.test = test
        self.kwargs = kwargs
        # Computed variables
        self.n_keys = len(keys)
        self.total_variable_share = 1 - sum(d_key_fixed_share.values())

    def __iter__(self):
        for x_simplex in iterate_simplex_grid(d=self.n_keys, denominator=self.denominator):
            x_simplex = np.array(x_simplex) * self.total_variable_share
            d_key_share = dict(zip(self.keys, x_simplex))
            for key, fixed_share in self.d_key_fixed_share.items():
                d_key_share[key] = d_key_share.get(key, 0) + fixed_share
            result = self.cls(d_key_share, **self.kwargs)
            if self.test is None or self.test(result):
                yield result



from fractions import Fraction
from poisson_approval import SimplexToProfile, ProfileNoisyDiscrete


def test_without_fixed_voters():
    """
        >>> simplex_to_profile = SimplexToProfile(
        ...     ProfileNoisyDiscrete, left_type=('abc', 0.5, 0.01), right_type=('bac', 0.5, 0.01), top_type='c>a~b')
        >>> profile = simplex_to_profile(left=Fraction(11, 80), top=Fraction(52, 80), right=Fraction(17, 80))
        >>> print(profile)
        <abc 0.5 ± 0.01: 11/80, bac 0.5 ± 0.01: 17/80, c>a~b: 13/20> (Condorcet winner: c)

    """
    pass


def test_profile_noisy_discrete_without_noise():
    """
        >>> simplex_to_profile = SimplexToProfile(
        ...     ProfileNoisyDiscrete, left_type=('abc', 0.5), right_type=('bac', 0.5), top_type='c>a~b', noise=0.01)
        >>> profile = simplex_to_profile(left=Fraction(11, 80), top=Fraction(52, 80), right=Fraction(17, 80))
        >>> print(profile)
        <abc 0.5 ± 0.01: 11/80, bac 0.5 ± 0.01: 17/80, c>a~b: 13/20> (Condorcet winner: c)
    """
    pass



RANKINGS = ['abc', 'acb', 'bac', 'bca', 'cab', 'cba']
"""list of str: All possible rankings, such as ``'abc'``."""

PERMUTATIONS = RANKINGS
"""list of str: Alias for ``RANKINGS``."""

TWELVE_TYPES = ['a_bc', 'ab_c', 'a_cb', 'ac_b', 'b_ac', 'ba_c', 'b_ca', 'bc_a', 'c_ab', 'ca_b', 'c_ba', 'cb_a']
"""
list of str: All possible twelve-types, such as ``'a_bc'`` or ``'ab_c'``. Cf. :class:`~poisson_approval.ProfileTwelve`.
"""

SETS_OF_RANKINGS_UP_TO_RELABELLING = [
    ('abc', ),
    ('abc', 'acb'),
    ('abc', 'bac'),
    ('abc', 'bca'),
    ('abc', 'cba'),
    ('abc', 'bac', 'cab'),
    ('abc', 'acb', 'bac'),
    ('abc', 'acb', 'bca'),
    ('abc', 'bca', 'cab'),
    ('abc', 'acb', 'bac', 'bca'),
    ('abc', 'acb', 'bac', 'cab'),
    ('abc', 'acb', 'bac', 'cba'),
    ('abc', 'acb', 'bca', 'cba'),
    ('abc', 'acb', 'bac', 'bca', 'cab'),
    ('abc', 'acb', 'bac', 'bca', 'cab', 'cba')
]
"""list of tuple of str: All possible sets of rankings, up to relabelling the candidates."""

WEAK_ORDERS_LOVE_WITHOUT_INVERSIONS = ['a>b~c', 'b>a~c', 'c>a~b']
"""
list of str: All possible weak orders of type `love`, without inversion. I.e. there is ``'a>b~c'`` but not the
equivalent ``'a>c~b'``.
"""

WEAK_ORDERS_HATE_WITHOUT_INVERSIONS = ['a~b>c', 'a~c>b', 'b~c>a']
"""
list of str: All possible weak orders of type `hate`, without inversion. I.e. there is ``'a~b>c'`` but not the
equivalent ``'b~a>c'``.
"""

WEAK_ORDERS_WITHOUT_INVERSIONS = WEAK_ORDERS_LOVE_WITHOUT_INVERSIONS + WEAK_ORDERS_HATE_WITHOUT_INVERSIONS
"""
list of str: All possible weak orders, without inversion. I.e. there is ``'a>b~c'`` but not the equivalent ``'a>c~b'``.
"""

CANDIDATES = ['a', 'b', 'c']
"""list of str: All candidates."""

PAIRS_WITHOUT_INVERSIONS = ['ab', 'ac', 'bc']
"""
list of str: All pairs of candidates, without inversion. I.e. there is ``'ab'`` but not the equivalent ``'ba'``.
"""

PAIRS_INVERTED = ['ba', 'ca', 'cb']
"""
list of str: All pairs of candidates in reverse alphabetical order. I.e. there is ``'ba'`` but not the equivalent
``'ab'``.
"""

PAIRS_WITH_INVERSIONS = PAIRS_WITHOUT_INVERSIONS + PAIRS_INVERTED
"""
list of str: All pairs of candidates, including inversions. I.e. there are both ``'ab'`` and ``'ba'``.
"""

BALLOTS_WITHOUT_INVERSIONS = CANDIDATES + PAIRS_WITHOUT_INVERSIONS
"""
list of str: All possible ballots, without inversions. I.e. there is ``'ab'`` but not the equivalent ``'ba'``.
"""

BALLOTS_WITH_INVERSIONS = BALLOTS_WITHOUT_INVERSIONS + PAIRS_INVERTED
"""
list of str: All possible ballots, with inversions. I.e. there are both ``'ab'`` and ``'ba'``.
"""

BALLOTS_WITHOUT_INVERSIONS_SORTED_ALPHABETICAL = sorted(BALLOTS_WITHOUT_INVERSIONS)
"""
list of str: All possible ballots, without inversions. I.e. there is ``'ab'`` but not the equivalent ``'ba'``.
The list is sorted alphabetically.
"""

UTILITY_DEPENDENT = 'utility-dependent'
SPLIT = 'Split'  # For 'haters' in Plurality or 'lovers' in Anti-Plurality
INCONCLUSIVE = 'inconclusive'

APPROVAL = 'Approval'
"""str: Approval."""

PLURALITY = 'Plurality'
"""str: Plurality."""

ANTI_PLURALITY = 'Anti-plurality'
"""str: Anti-Plurality."""

VOTING_RULES = [APPROVAL, PLURALITY, ANTI_PLURALITY]
"""list of str: The three voting rules of the package, i.e. Approval, Plurality and Anti-Plurality."""


def _f_abc_xyz(my_list):
    return [element.replace('a', 'x').replace('b', 'y').replace('c', 'z') for element in my_list]


XYZ_RANKINGS = _f_abc_xyz(RANKINGS)
XYZ_PERMUTATIONS = XYZ_RANKINGS
XYZ_TWELVE_TYPES = _f_abc_xyz(TWELVE_TYPES)
XYZ_BALLOTS_WITHOUT_INVERSION = _f_abc_xyz(BALLOTS_WITHOUT_INVERSIONS)
XYZ_BALLOTS_INVERTED = _f_abc_xyz(PAIRS_INVERTED)
XYZ_BALLOTS_WITH_INVERSIONS = _f_abc_xyz(BALLOTS_WITH_INVERSIONS)
XYZ_WEAK_ORDERS_WITHOUT_INVERSIONS = _f_abc_xyz(WEAK_ORDERS_WITHOUT_INVERSIONS)


NORMALIZATION_WARNING = ("Warning: the input is not normalized, I will normalize it. If you use floats, consider using "
                         "fractions. To disable this warning, use the parameter normalization_warning=False.")



import pytest
from poisson_approval import ternary_figure


def test():
    """
        >>> def f(right, top, left):
        ...     return (right**2 + top) / (left + 1)
        >>> figure, tax = ternary_figure(scale=10, size_inches=(6, 4))
        >>> tax.heatmap_intensity(f,
        ...                       left_label='left',
        ...                       right_label='right',
        ...                       top_label='top',
        ...                       cb_kwargs={})
        >>> tax.set_title_padded('An intensity heat map')
    """
    pass


def test_save_data():
    """
        >>> def g(right, top, left):
        ...     a = top**.5
        ...     b = left**2
        ...     c = 1 - a - b
        ...     return [a, b, c]
        >>> figure, tax = ternary_figure(scale=5)
        >>> tax.heatmap_candidates(g,
        ...                        left_label='left',
        ...                        right_label='right',
        ...                        top_label='top',
        ...                        legend_title='Candidates',
        ...                        legend_style='palette',
        ...                        file_save_data='zzz_test_ternary_plots.sav')
        >>> tax.set_title_padded('A candidate heat map')
    """
    pass


def test_f_point_values_():
    with pytest.raises(ValueError):
        figure, tax = ternary_figure(scale=5)
        tax.f_point_values_(right=0.5, top=0.3, left=0.2)


def test_annotate_condorcet_old():
    figure, tax = ternary_figure(scale=5)
    with pytest.raises(NotImplementedError):
        tax._annotate_condorcet_old(right_ranking='abc', left_ranking='bac', top_ranking='cab',
                                    d_order_fixed_share={'cba': 0.2})
    # Same top
    tax._annotate_condorcet_old(right_ranking='abc', left_ranking='abc', top_ranking='acb')
    # Two different tops
    tax._annotate_condorcet_old(right_ranking='abc', left_ranking='acb', top_ranking='bac')
    tax._annotate_condorcet_old(right_ranking='abc', left_ranking='bac', top_ranking='acb')
    tax._annotate_condorcet_old(right_ranking='bac', left_ranking='abc', top_ranking='acb')
    # Three different tops
    # Sub-case: Three different bottoms (Condorcet cycle)
    tax._annotate_condorcet_old(right_ranking='abc', left_ranking='bca', top_ranking='cab')
    # Sub-case: Two different bottoms (divided majority)
    tax._annotate_condorcet_old(right_ranking='abc', left_ranking='bac', top_ranking='cab')
    tax._annotate_condorcet_old(right_ranking='bac', left_ranking='abc', top_ranking='cab')
    tax._annotate_condorcet_old(right_ranking='bac', left_ranking='cab', top_ranking='abc')



"""Top-level package for Poisson Approval."""

__author__ = """François Durand"""
__email__ = 'fradurand@gmail.com'
__version__ = '0.29.0'


# Utils
from poisson_approval.utils.computation_engine import computation_engine
from poisson_approval.utils.ComputationEngine import ComputationEngine
from poisson_approval.utils.ComputationEngineNumeric import ComputationEngineNumeric
from poisson_approval.utils.ComputationEngineSymbolic import ComputationEngineSymbolic
from poisson_approval.utils.DictPrintingInOrder import DictPrintingInOrder
from poisson_approval.utils.DictPrintingInOrderIgnoringNone import DictPrintingInOrderIgnoringNone
from poisson_approval.utils.DictPrintingInOrderIgnoringZeros import DictPrintingInOrderIgnoringZeros
from poisson_approval.utils.SetPrintingInOrder import SetPrintingInOrder
from poisson_approval.utils.Util import initialize_random_seeds, rand_simplex, rand_integers_fixed_sum, \
    rand_simplex_grid, probability, image_distribution, isnan, isposinf, isneginf, give_figure, to_callable, \
    product_dict, candidates_to_d_candidate_probability, candidates_to_probabilities, array_to_d_candidate_value, \
    d_candidate_value_to_array, one_over_t, one_over_sqrt_t, one_over_log_t_plus_one, \
    one_over_log_log_t_plus_fourteen, my_division, iterator_integers_fixed_sum, iterate_simplex_grid
from poisson_approval.utils.UtilBallots import ballot_one, ballot_two, ballot_one_two, ballot_one_three, \
    ballot_high_u, ballot_low_u, allowed_ballots
from poisson_approval.utils.UtilCache import cached_property, DeleteCacheMixin, property_deleting_cache
from poisson_approval.utils.UtilMasks import masks_area_naive, masks_area, masks_distribution_naive, \
    masks_distribution, winners_distribution, random_mask, random_masks
from poisson_approval.utils.UtilPlot import plt_cdf, plt_step_with_error, plt_plot_with_error
from poisson_approval.utils.UtilPreferences import is_hater, is_lover, is_weak_order, sort_weak_order

# Constants
from poisson_approval.constants.basic_constants import *
from poisson_approval.constants.EquilibriumStatus import EquilibriumStatus
from poisson_approval.constants.Focus import Focus

# Simple containers
from poisson_approval.containers.Winners import Winners
from poisson_approval.containers.Scores import Scores
from poisson_approval.containers.AnalyzedStrategies import AnalyzedStrategies

# Events
from poisson_approval.events.Asymptotic import Asymptotic
from poisson_approval.events.Event import Event
from poisson_approval.events.EventDuo import EventDuo
from poisson_approval.events.EventPivotStrict import EventPivotStrict
from poisson_approval.events.EventPivotTij import EventPivotTij
from poisson_approval.events.EventPivotTjk import EventPivotTjk
from poisson_approval.events.EventPivotWeak import EventPivotWeak
from poisson_approval.events.EventTrio import EventTrio
from poisson_approval.events.EventTrio1t import EventTrio1t
from poisson_approval.events.EventTrio2t import EventTrio2t

# Best response
from poisson_approval.best_response.BestResponse import BestResponse
from poisson_approval.best_response.BestResponseAntiPlurality import BestResponseAntiPlurality
from poisson_approval.best_response.BestResponseApproval import BestResponseApproval
from poisson_approval.best_response.BestResponsePlurality import BestResponsePlurality

# Tau-vector
from poisson_approval.tau_vector.TauVector import TauVector

# Strategies
from poisson_approval.strategies.Strategy import Strategy
from poisson_approval.strategies.StrategyTwelve import StrategyTwelve
from poisson_approval.strategies.StrategyThreshold import StrategyThreshold
from poisson_approval.strategies.StrategyOrdinal import StrategyOrdinal

# Profiles
from poisson_approval.profiles.Profile import Profile
from poisson_approval.profiles.ProfileDiscrete import ProfileDiscrete
from poisson_approval.profiles.ProfileNoisyDiscrete import ProfileNoisyDiscrete
from poisson_approval.profiles.ProfileOrdinal import ProfileOrdinal
from poisson_approval.profiles.ProfileCardinal import ProfileCardinal
from poisson_approval.profiles.ProfileCardinalContinuous import ProfileCardinalContinuous
from poisson_approval.profiles.ProfileTwelve import ProfileTwelve
from poisson_approval.profiles.ProfileHistogram import ProfileHistogram

# Iterables
from poisson_approval.iterables.IterableProfileDiscreteGrid import IterableProfileDiscreteGrid
from poisson_approval.iterables.IterableProfileHistogramGrid import IterableProfileHistogramGrid
from poisson_approval.iterables.IterableProfileNoisyDiscreteGrid import IterableProfileNoisyDiscreteGrid
from poisson_approval.iterables.IterableProfileOrdinalGrid import IterableProfileOrdinalGrid
from poisson_approval.iterables.IterableProfileTwelveGrid import IterableProfileTwelveGrid
from poisson_approval.iterables.IterableSimplexGrid import IterableSimplexGrid
from poisson_approval.iterables.IterableStrategyOrdinal import IterableStrategyOrdinal
from poisson_approval.iterables.IterableStrategyThresholdGrid import IterableStrategyThresholdGrid
from poisson_approval.iterables.IterableStrategyTwelve import IterableStrategyTwelve
from poisson_approval.iterables.IterableTauVectorGrid import IterableTauVectorGrid

# Random factories
from poisson_approval.random_factories.RandConditional import RandConditional
from poisson_approval.random_factories.RandProfileDiscreteGridUniform import RandProfileDiscreteGridUniform
from poisson_approval.random_factories.RandProfileDiscreteUniform import RandProfileDiscreteUniform
from poisson_approval.random_factories.RandProfileHistogramGridUniform import RandProfileHistogramGridUniform
from poisson_approval.random_factories.RandProfileHistogramUniform import RandProfileHistogramUniform
from poisson_approval.random_factories.RandProfileNoisyDiscreteGridUniform import RandProfileNoisyDiscreteGridUniform
from poisson_approval.random_factories.RandProfileNoisyDiscreteUniform import RandProfileNoisyDiscreteUniform
from poisson_approval.random_factories.RandProfileOrdinalGridUniform import RandProfileOrdinalGridUniform
from poisson_approval.random_factories.RandProfileOrdinalUniform import RandProfileOrdinalUniform
from poisson_approval.random_factories.RandProfileTwelveGridUniform import RandProfileTwelveGridUniform
from poisson_approval.random_factories.RandProfileTwelveUniform import RandProfileTwelveUniform
from poisson_approval.random_factories.RandSimplexGridUniform import RandSimplexGridUniform
from poisson_approval.random_factories.RandSimplexUniform import RandSimplexUniform
from poisson_approval.random_factories.RandStrategyOrdinalUniform import RandStrategyOrdinalUniform
from poisson_approval.random_factories.RandStrategyThresholdGridUniform import RandStrategyThresholdGridUniform
from poisson_approval.random_factories.RandStrategyThresholdUniform import RandStrategyThresholdUniform
from poisson_approval.random_factories.RandStrategyTwelveUniform import RandStrategyTwelveUniform
from poisson_approval.random_factories.RandTauVectorGridUniform import RandTauVectorGridUniform
from poisson_approval.random_factories.RandTauVectorUniform import RandTauVectorUniform

# Meta-analysis
from poisson_approval.meta_analysis.NiceStatsProfileOrdinal import NiceStatsProfileOrdinal
from poisson_approval.meta_analysis.binary_plots import BinaryAxesSubplotPoisson, binary_figure
from poisson_approval.meta_analysis.binary_shortcuts import binary_plot_n_equilibria, \
    binary_plot_winners_at_equilibrium, binary_plot_winning_frequencies, binary_plot_convergence, \
    XyyToProfile
from poisson_approval.meta_analysis.convergence_test import convergence_test
from poisson_approval.meta_analysis.is_condorcet import is_condorcet
from poisson_approval.meta_analysis.is_not_condorcet import is_not_condorcet
from poisson_approval.meta_analysis.monte_carlo_fictitious_play import monte_carlo_fictitious_play, \
    MCS_BALLOT_STATISTICS, MCS_CONVERGES, MCS_DECREASING_SCORES, \
    MCS_FREQUENCY_CW_WINS, MCS_PROFILE, MCS_TAU_INIT, MCS_WELFARE_LOSSES, \
    MCS_UTILITY_THRESHOLDS, MCS_CANDIDATE_WINNING_FREQUENCY, MCS_N_EPISODES
from poisson_approval.meta_analysis.plot_welfare_losses import plot_welfare_losses
from poisson_approval.meta_analysis.plot_distribution_scores import plot_distribution_scores
from poisson_approval.meta_analysis.plot_utility_thresholds import plot_utility_thresholds
from poisson_approval.meta_analysis.ternary_plots import TernaryAxesSubplotPoisson, ternary_figure
from poisson_approval.meta_analysis.ternary_shortcuts import ternary_plot_n_equilibria, \
    ternary_plot_winners_at_equilibrium, ternary_plot_winning_frequencies, ternary_plot_convergence, \
    SimplexToProfile



from poisson_approval.constants.basic_constants import CANDIDATES, PAIRS_WITHOUT_INVERSIONS
from poisson_approval.utils.Util import my_division, my_sign
from poisson_approval.utils.UtilPreferences import d_candidate_ordinal_utility


def _intersection(x, y):
    """Intersection of two multi-intervals.

    Parameters
    ----------
    x, y : list
        Each parameter represents a multi-interval. E.g. [1, 2, 12, 51] represents the union of the interval [1, 2]
        and the interval [12, 51]. All bounds are assumed sorted.

    Returns
    -------
    list
        The intersection, which is itself a multi-interval.

    Examples
    --------
        >>> x = [1, 6, 7, 10]
        >>> y = [0, 2, 3, 4, 5, 8, 9, 11]
        >>> _intersection(x, y)
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    """
    events = [(val, i % 2 == 0) for i, val in enumerate(x)]
    events.extend([(val, i % 2 == 0) for i, val in enumerate(y)])
    events.sort()
    result = []
    active = 0
    for val, b in events:
        if active == 2:
            result.append(val)
            active -= 1
        else:
            if b:
                active += 1
            else:
                active -= 1
            if active == 2:
                result.append(val)
    return result


def _difference(x, y):
    """Difference of two multi-intervals.

    Parameters
    ----------
    x, y : list
        Each parameter represents a multi-interval. E.g. [1, 2, 12, 51] represents the union of the interval [1, 2]
        and the interval [12, 51]. All bounds are assumed sorted.

    Returns
    -------
    list
        The difference (in a set-theory sense), which is itself a multi-interval.

    Examples
    --------
        >>> x = [1, 6, 7, 10]
        >>> y = [0, 2, 3, 4, 5, 8, 9, 11]
        >>> _difference(x, y)
        [2, 3, 4, 5, 8, 9]
    """
    events = [(val, i % 2 == 0) for i, val in enumerate(x)]
    events.extend([(val, i % 2 == 1) for i, val in enumerate(y)])
    events.sort()
    result = []
    active = 1
    for val, b in events:
        if active == 2:
            result.append(val)
            active -= 1
        else:
            if b:
                active += 1
            else:
                active -= 1
            if active == 2:
                result.append(val)
    return result


def _interval_victory(x, y, plus=0, zero=0, minus=0):
    """Interval of victory.

    Interval of victory of one candidate, say `a`, against another candidate, say `b`.

    Parameters
    ----------
    x, y : Number in {-1, 0, 1}
        For example, `x` is 1 if the first order likes `a` better than `b`, 0 if she is indifferent between them,
        and -1 if she likes `b` better than `a`. The role of `y` is similar for the second order.
    plus : Number in [0, 1].
        Fixed share of voters who like `a` better than `b`.
    zero : Number in [0, 1].
        Fixed share of voters who are indifferent between `a` and `b`.
    minus : Number in [0, 1].
        Fixed share of voters who like `a` less than `b`.

    Returns
    -------
    list
        The interval of victory of `a` against `b`, in the setting defined as:

        * Fixed shares of voters `plus`, `zero` and `minus` like respectively `a` better than, as much as or less
          than `b`,
        * The remaining voters are split between the first and the second orders in proportions that are given
          by the point on the interval [0, 1].

        If the interval is empty, an empty list is returned.

    Examples
    --------
        >>> _interval_victory(-1, 1, minus=0.2)
        [0.625, 1]
        >>> _interval_victory(1, 1)
        [0, 1]
        >>> _interval_victory(-1, -1)
        []
        >>> _interval_victory(1, 0, minus=0.5)
        []
    """
    v = 1 - plus - zero - minus  # Variable share (for variables voters)
    adv = plus - minus  # Constant advantage for `a`
    if x == y:
        net_result = adv + v * x
        if net_result >= 0:
            return [0, 1]
        else:
            return []
    else:
        beta = my_division(x + my_division(adv, v), x - y)
        beta = min(max(beta, 0), 1)
        if x > y:  # Left is more favorable to `a`
            result = [0, beta]
        else:
            result = [beta, 1]
        if result[0] == result[1]:
            return []
        else:
            return result


def _interval_condorcet(candidate, order_x, order_y, d_order_fixed_share=None):
    """Interval representing the zone where a candidate is a Condorcet winner.

    Parameters
    ----------
    candidate : str
        'a', 'b' or 'c'.
    order_x, order_y : str
        Rankings ('abc') or weak orders ('a~b>c' or 'a>b~c').
    d_order_fixed_share : dict, optional
        Key: order (ranking or weak order). Value: a fixed share of voters in [0, 1].

    Returns
    -------
    list
        Interval where `candidate` is the CW, in the setting defined as:

        * Fixed shares of voters `plus`, `zero` and `minus` like respectively `a` better than, as much as or less
          than `b`,
        * The remaining voters are split between the first and the second orders in proportions that are given
          by the point on the interval [0, 1].

        If the interval is empty, an empty list is returned.

    Examples
    --------
        >>> _interval_condorcet('a', 'abc', 'cba', {'bac': 0.2})
        [0, 0.375]
        >>> _interval_condorcet('b', 'abc', 'cba', {'bac': 0.2})
        [0.375, 0.625]
        >>> _interval_condorcet('c', 'abc', 'cba', {'bac': 0.2})
        [0.625, 1]
        >>> _interval_condorcet('c', 'abc', 'cba', {'b>a~c': 0.2})
        [0.625, 1]
    """
    d_order_fixed_share = dict() if d_order_fixed_share is None else d_order_fixed_share
    other_candidates = sorted({'a', 'b', 'c'} - {candidate})
    d_candidate_pseudo_utility_x = d_candidate_ordinal_utility(order_x)
    d_candidate_pseudo_utility_y = d_candidate_ordinal_utility(order_y)
    d_candidate_pseudo_utilities_fixed = [
        d_candidate_ordinal_utility(order) for order, share in d_order_fixed_share.items()]
    interval = [0, 1]
    for other in other_candidates:
        plus = 0
        zero = 0
        minus = 0
        for d_candidate_pseudo_utility, share in zip(d_candidate_pseudo_utilities_fixed, d_order_fixed_share.values()):
            if d_candidate_pseudo_utility[candidate] > d_candidate_pseudo_utility[other]:
                plus += share
            elif d_candidate_pseudo_utility[candidate] == d_candidate_pseudo_utility[other]:
                zero += share
            else:
                minus += share
        interval_victory = _interval_victory(
            my_sign(d_candidate_pseudo_utility_x[candidate] - d_candidate_pseudo_utility_x[other]),
            my_sign(d_candidate_pseudo_utility_y[candidate] - d_candidate_pseudo_utility_y[other]),
            plus, zero, minus
        )
        interval = _intersection(interval, interval_victory)
    return interval


def _draw_interval(ax, interval, name):
    """Draw a multi-interval in a binary plot.

    Parameters
    ----------
    ax : Axes
    interval : list
        The multi-interval.
    name : str
        The name to display in the center of the interval.
    """
    i = 0
    while i < len(interval):
        inf = interval[i]
        sup = interval[i + 1]
        i += 2
        if inf > 0:
            ax.axvline(x=inf, ymin=0, ymax=1, color='black')
        if sup < 1:
            ax.axvline(x=sup, ymin=0, ymax=1, color='black')
        if inf < sup:
            ax.annotate(name, ((inf + sup) / 2, 0.5), verticalalignment='center', horizontalalignment='center')


def draw_condorcet_intervals(ax, left_ranking, right_ranking, d_order_fixed_share=None):
    """Draw and annotate the Condorcet intervals.

    Cf. :meth:`BinaryAxesSubplotPoisson.annotate_condorcet`.
    """
    interval_not_condorcet = (0, 1)
    d_candidate_interval = dict()
    for candidate in CANDIDATES:
        d_candidate_interval[candidate] = _interval_condorcet(
            candidate, left_ranking, right_ranking, d_order_fixed_share)
        interval_not_condorcet = _difference(interval_not_condorcet, d_candidate_interval[candidate])
    _draw_interval(ax, interval_not_condorcet, 'no CW')
    # Three candidates
    interval_abc = _intersection(_intersection(d_candidate_interval['a'], d_candidate_interval['b']),
                                 d_candidate_interval['c'])
    _draw_interval(ax, interval_abc, 'a, b, c are CW')
    for candidate in CANDIDATES:
        d_candidate_interval[candidate] = _difference(d_candidate_interval[candidate], interval_abc)
    # Two candidates
    d_pair_interval = dict()
    for c1, c2 in PAIRS_WITHOUT_INVERSIONS:
        d_pair_interval[c1 + c2] = _intersection(d_candidate_interval[c1], d_candidate_interval[c2])
        _draw_interval(ax, d_pair_interval[c1 + c2], '%s, %s are CW' % (c1, c2))
    for c1, c2 in PAIRS_WITHOUT_INVERSIONS:
        d_candidate_interval[c1] = _difference(d_candidate_interval[c1], d_pair_interval[c1 + c2])
        d_candidate_interval[c2] = _difference(d_candidate_interval[c2], d_pair_interval[c1 + c2])
    # One candidate
    for candidate in CANDIDATES:
        _draw_interval(ax, d_candidate_interval[candidate], '%s is CW' % candidate)



from poisson_approval import NiceStatsProfileOrdinal, initialize_random_seeds, ProfileOrdinal


def test():
    # Test when rejection is indeed necessary.
    initialize_random_seeds()
    nice_stats = NiceStatsProfileOrdinal(
        tests_profile=[
            (lambda profile: any([strategy for strategy in profile.analyzed_strategies_ordinal.equilibria
                                  if strategy.profile.condorcet_winners == strategy.winners]),
             'There exists a true equilibrium electing the CW'),
        ],
        conditional_on=lambda profile: profile.has_majority_ranking)
    nice_stats.run(n_samples=2)
    profile = nice_stats.find_example(0, False)
    assert isinstance(profile, ProfileOrdinal)


def test_misc_options():
    """
        >>> initialize_random_seeds()
        >>> nice_stats = NiceStatsProfileOrdinal(
        ...     tests_profile=[
        ...         (lambda profile: any([strategy for strategy in profile.analyzed_strategies_ordinal.equilibria
        ...                               if strategy.profile.condorcet_winners == strategy.winners]),
        ...          'There exists a true equilibrium electing the CW'),
        ...     ],
        ...     tests_strategy=[
        ...         (lambda strategy: strategy.profile.condorcet_winners == strategy.winners,
        ...          'There exists an equilibrium that elects the CW')
        ...     ],
        ...     tests_strategy_dist=[
        ...         (lambda strategy: strategy.profile.condorcet_winners == strategy.winners,
        ...          'There exists an equilibrium that elects the CW')
        ...     ],
        ...     tests_strategy_winners = [
        ...         (lambda sigma: True, 'Number of possible winners')
        ...     ],
        ...     conditional_on=lambda profile: profile.is_profile_condorcet == 1.
        ... )
        >>> nice_stats.run(n_samples=10)
        >>> nice_stats.plot_test_strategy(test=0, ylabel=False, legend=True, replacement_name='foobar')
        >>> nice_stats.plot_cutoff(test=0)
    """
    pass



class SuperclassMeta(type):
    """
    Meta-class for superclasses.

    If a member of a subclass doesn't have a docstring, it copies the docstring of the corresponding member of
    the superclass.

    Source: https://stackoverflow.com/questions/40508492/python-sphinx-inherit-method-documentation-from-superclass.
    """

    def __new__(mcs, classname, bases, cls_dict):
        cls = super().__new__(mcs, classname, bases, cls_dict)
        for name, member in cls_dict.items():
            if not getattr(member, '__doc__'):
                try:
                    member.__doc__ = getattr(bases[-1], name).__doc__
                except (AttributeError, IndexError):
                    pass
        return cls



import numpy as np
from matplotlib import pyplot as plt
from poisson_approval.constants.basic_constants import VOTING_RULES
from poisson_approval.utils.UtilPlot import plt_cdf
from poisson_approval.random_factories.RandProfileHistogramUniform import RandProfileHistogramUniform
from poisson_approval.meta_analysis.monte_carlo_fictitious_play import monte_carlo_fictitious_play, \
    MCS_WELFARE_LOSSES


def plot_welfare_losses(results, criterion, **kwargs):
    """
    Plot the distribution (CDF) of the welfare losses, for each voting rule.

    Parameters
    ----------
    results : dict
        Results of :func:`~poisson_approval.meta_analysis.monte_carlo_fictitious_play.monte_carlo_fictitious_play`,
        with at least the setting
        :const:`~poisson_approval.meta_analysis.monte_carlo_fictitious_play.MCS_WELFARE_LOSSES`.
    criterion : str
        ``'utilitarian_welfare_losses'``, ``'plurality_welfare_losses'`` or ``'anti_plurality_welfare_losses'``.
    kwargs
        Other keyword arguments are passed to the function ``step`` of `matplotlib`.

    Examples
    --------
        >>> meta_results = monte_carlo_fictitious_play(
        ...     factory=RandProfileHistogramUniform(n_bins=1),
        ...     n_samples=1,
        ...     n_max_episodes=10,
        ...     voting_rules=VOTING_RULES,
        ...     monte_carlo_settings=[MCS_WELFARE_LOSSES],
        ... )
        >>> plot_welfare_losses(meta_results, 'utilitarian_welfare_losses')
        >>> plot_welfare_losses(meta_results, 'plurality_welfare_losses')
        >>> plot_welfare_losses(meta_results, 'anti_plurality_welfare_losses')
    """
    fig, ax = plt.subplots(figsize=(8, 4))
    voting_rules = results.keys()
    for voting_rule in voting_rules:
        n_samples = results[voting_rule]['n_samples']
        welfare_losses = np.array(results[voting_rule][criterion]).flatten()
        weights = np.array(results[voting_rule]['candidate_winning_frequencies']).flatten() / n_samples
        plt_cdf(welfare_losses, weights, n_samples, label=voting_rule, **kwargs)
    ax.grid(True)
    ax.legend()
    if criterion == 'utilitarian_welfare_losses':
        ax.set_xlabel('Loss of utilitarian welfare')
    elif criterion == 'plurality_welfare_losses':
        ax.set_xlabel('Loss of anti-Rawlsian welfare')
    elif criterion == 'anti_plurality_welfare_losses':
        ax.set_xlabel('Loss of Rawlsian welfare')
    else:  # pragma: no cover - Should never happen
        ax.set_xlabel('Loss of welfare')
    ax.set_ylabel('Cumulative likelihood of occurrence')
    plt.xlim(0, 1)
    plt.xticks(np.arange(0, 1.1, 0.1))



import pytest
from poisson_approval import allowed_ballots


def test_allowed_ballots():
    with pytest.raises(NotImplementedError):
        allowed_ballots(voting_rule='Non implemented voting rule')



from poisson_approval import convergence_test, ProfileHistogram


def test_convergence_test_iterated_voting():
    """
        >>> converges = convergence_test(
        ...     n_max_episodes=10,
        ...     perception_update_ratio=1,
        ...     ballot_update_ratio=1,
        ...     meth='iterated_voting',
        ... )
        >>> profile = ProfileHistogram({'abc': 0.7, 'bac': 0.3}, {'abc': [1.], 'bac': [1.]})
        >>> converges(profile)
        True
    """
    pass



from poisson_approval.constants.basic_constants import *
from poisson_approval.iterables.IterableSimplexGrid import IterableSimplexGrid
from poisson_approval.profiles.ProfileOrdinal import ProfileOrdinal


class IterableProfileOrdinalGrid:
    """Iterate over ordinal profiles (:class:`ProfileOrdinal`) defined on a grid.

    Parameters
    ----------
    denominator : int or iterable
        The grain(s) of the grid.
    orders : iterable, optional
        These orders will have a variable share. They can be rankings, e.g. ``'abc'``, or weak orders, e.g.
        ``'a~b>c'``. Default: all rankings.
    d_order_fixed_share : dict, optional
        A dictionary. For each entry ``order: fixed_share``, this order will have at least this fixed share. The total
        must be lower or equal to 1.
    standardized : bool, optional
        If True, then only standardized profiles are given. Cf. :meth:`Profile.is_standardized`. You should
        probably use this option if the arguments `orders`, `d_order_fixed_share` and `test` treat the candidates
        symmetrically.
    test : callable, optional
        A function ``ProfileOrdinal -> bool``. Only profiles meeting this test are given.
    kwargs
        Additional parameters are passed to :class:`ProfileOrdinal` when creating the profile.

    Examples
    --------
    Basic usage:

        >>> for profile in IterableProfileOrdinalGrid(denominator=3, standardized=True):
        ...     print(profile)
        <abc: 1> (Condorcet winner: a)
        <abc: 2/3, acb: 1/3> (Condorcet winner: a)
        <abc: 2/3, bac: 1/3> (Condorcet winner: a)
        <abc: 2/3, bca: 1/3> (Condorcet winner: a)
        <abc: 2/3, cab: 1/3> (Condorcet winner: a)
        <abc: 2/3, cba: 1/3> (Condorcet winner: a)
        <abc: 1/3, acb: 1/3, bac: 1/3> (Condorcet winner: a)
        <abc: 1/3, acb: 1/3, bca: 1/3> (Condorcet winner: a)
        <abc: 1/3, bac: 1/3, cab: 1/3> (Condorcet winner: a)
        <abc: 1/3, bca: 1/3, cab: 1/3>

    For more examples, cf. :class:`IterableSimplexGrid`.
    """
    def __init__(self, denominator, orders=None, d_order_fixed_share=None, standardized=False, test=None, **kwargs):
        if orders is None:
            orders = RANKINGS
        self.standardized = standardized
        self._base_iterator = IterableSimplexGrid(cls=ProfileOrdinal, denominator=denominator, keys=orders,
                                                  d_key_fixed_share=d_order_fixed_share, test=test, **kwargs)

    def __iter__(self):
        return (profile for profile in self._base_iterator
                if not self.standardized or profile.is_standardized)



from poisson_approval.best_response.BestResponse import BestResponse
from poisson_approval.constants.basic_constants import *
from poisson_approval.utils.UtilCache import cached_property


class BestResponsePlurality(BestResponse):
    """Best response for a given ordinal type of voter in Plurality.

    The main objective of this class is to compute :attr:`utility_threshold`.

    For the sake of consistency with :class:`BestResponseApproval`, it provides the string :attr:`justification`,
    indicating which sub-algorithm was used. But since there are no actual sub-algorithms for plurality, the
    justification is always the same: ``'Plurality analysis'``.

    Parameters
    ----------
    tau : TauVector
        A tau-vector.
    ranking : str
        Voter's ranking, e.g. ``'abc'``.

    Attributes
    ----------
    i, j, k : str
        The first (resp. second, third) candidate in `ranking`. E.g. ``a``.
    ij, ik, jk : str
        The ballots with two candidates. E.g. ``ab``.
    tau_i, tau_j, tau_k, tau_ij, tau_ik, tau_jk : Number
        The values of the tau-vector.
    """

    PLURALITY_ANALYSIS = 'Plurality analysis'
    voting_rule = PLURALITY

    @cached_property
    def results(self):
        """tuple : Tuple `(utility_threshold, justification)`. Cf. :attr:`utility_threshold` and :attr:`justification`.
        """
        assert self.tau.voting_rule == PLURALITY
        if self.tau_i < self.tau_j and self.tau_i < self.tau_k:
            # The best response is `j`.
            utility_threshold = self.ce.S(0)
        elif 0 < self.tau_i == self.tau_k < self.tau_j:
            utility_threshold = self.ce.Rational(1, 2)
        else:
            # The best response is `i`.
            utility_threshold = self.ce.S(1)
        return utility_threshold, self.PLURALITY_ANALYSIS



from poisson_approval.profiles.ProfileHistogram import ProfileHistogram
from poisson_approval.utils.Util import one_over_log_t_plus_one


def convergence_test(n_max_episodes, init='sincere',
                     perception_update_ratio=one_over_log_t_plus_one,
                     ballot_update_ratio=one_over_log_t_plus_one,
                     meth='fictitious_play'):
    """
    Create a convergence test.

    Parameters
    ----------
    n_max_episodes : int
        Maximum number of episodes for the fictitious play / iterated voting.
    init : Strategy or TauVector or str
        Cf. :meth:`~poisson_approval.ProfileCardinal.fictitious_play`
        or :meth:`~poisson_approval.ProfileCardinal.iterated_voting`.
    perception_update_ratio,ballot_update_ratio : callable or Number
        Cf. :meth:`~poisson_approval.ProfileCardinal.fictitious_play`
        or :meth:`~poisson_approval.ProfileCardinal.iterated_voting`.
    meth : str
        The name of the method (``'fictitious_play'`` or ``'iterated_voting'``).

    Returns
    -------
    callable
        A callable. Input: a profile. Output: a bool, which is True if the fictitious play / iterated
        voting converges (within the specified number of episodes).

    Examples
    --------
        >>> converges = convergence_test(n_max_episodes=10)
        >>> profile = ProfileHistogram({'abc': 0.7, 'bac': 0.3}, {'abc': [1.], 'bac': [1.]})
        >>> converges(profile)
        False

    This function is especially convenient when a test function is required, for example when using
    :func:`~poisson_approval.utils.Util.probability`.
    """
    def converges(profile):
        results = getattr(profile, meth)(init=init, n_max_episodes=n_max_episodes,
                                         perception_update_ratio=perception_update_ratio,
                                         ballot_update_ratio=ballot_update_ratio)
        return results['converges']
    return converges



from poisson_approval.constants.basic_constants import *
from poisson_approval.iterables.IterableSimplexGrid import IterableSimplexGrid
from poisson_approval.tau_vector.TauVector import TauVector
from poisson_approval.utils.UtilBallots import allowed_ballots


class IterableTauVectorGrid:
    """Iterate over tau-vectors (:class:`TauVector`) defined on a grid.

    Parameters
    ----------
    denominator : int or iterable
        The grain(s) of the grid.
    ballots : iterable, optional
        These ballots (e.g. ``'a'``, ``'ab'``) will have a variable share. Default: all allowed ballots.
    d_ballot_fixed_share : dict, optional
        A dictionary. For each entry ``ballot: fixed_share``, this ballot will have at least this fixed share. The total
        must be lower or equal to 1.
    standardized : bool, optional
        If True, then only standardized tau-vectors are given. Cf. :meth:`TauVector.is_standardized`. You should
        probably use this option if the arguments `ballots`, `d_ballot_fixed_share` and `test` treat the candidates
        symmetrically.
    test : callable, optional
        A function ``TauVector -> bool``. Only tau-vector meeting this test are given.
    kwargs
        Additional parameters are passed to :class:`TauVector` when creating the tau-vector.

    Examples
    --------
    Basic usage:

        >>> for tau in IterableTauVectorGrid(denominator=2, standardized=True):
        ...     print(tau)
        <a: 1> ==> a
        <a: 1/2, ab: 1/2> ==> a
        <a: 1/2, b: 1/2> ==> a, b
        <a: 1/2, bc: 1/2> ==> a, b, c
        <ab: 1> ==> a, b
        <ab: 1/2, ac: 1/2> ==> a

    If the voting rule is not approval, only the relevant ballots are used:

        >>> for tau in IterableTauVectorGrid(denominator=3, standardized=True, voting_rule=PLURALITY):
        ...     print(tau)
        <a: 1> ==> a (Plurality)
        <a: 2/3, b: 1/3> ==> a (Plurality)
        <a: 1/3, b: 1/3, c: 1/3> ==> a, b, c (Plurality)

    For more examples, cf. :class:`IterableSimplexGrid`.
    """
    def __init__(self, denominator, ballots=None, d_ballot_fixed_share=None, standardized=False, test=None, **kwargs):
        if ballots is None:
            try:
                ballots = allowed_ballots(kwargs['voting_rule'])
            except KeyError:
                ballots = allowed_ballots()
        self.standardized = standardized
        self._base_iterator = IterableSimplexGrid(cls=TauVector, denominator=denominator, keys=ballots,
                                                  d_key_fixed_share=d_ballot_fixed_share, test=test, **kwargs)

    def __iter__(self):
        return (tau for tau in self._base_iterator
                if not self.standardized or tau.is_standardized)



import numpy as np
from poisson_approval.constants.basic_constants import *
from poisson_approval.strategies.Strategy import Strategy
from poisson_approval.strategies.StrategyThreshold import StrategyThreshold
from poisson_approval.utils.Util import initialize_random_seeds


class RandStrategyThresholdUniform:
    """A random factory of threshold strategies (:class:`StrategyThreshold`) following the uniform distribution.

    Parameters
    ----------
    profile : Profile, optional
        The attached profile.
    voting_rule : str, optional
        The voting rule. Possible values are ``APPROVAL``, ``PLURALITY`` and ``ANTI_PLURALITY``. Default: the same
        voting rule as `profile` if a profile is specified, ``APPROVAL`` otherwise.
    d_ranking_fixed_strategy : dict
        Key: ranking. Value: fixed strategy. Cf. examples below.
    ratio_optimistic : Number, optional
        If specified, it will be applied to all voters, except those in `d_ranking_fixed_strategy`.
        If not specified, it depends on `profile`. If `profile` is specified and is an instance of
        :class:`ProfileCardinalContinuous`, then the ratios of optimistic voters are not mentioned in the strategy
        (because they are useless). In other cases, they are drawn at random.

    Examples
    --------
    Basic usage:

        >>> initialize_random_seeds()
        >>> rand_strategy = RandStrategyThresholdUniform()
        >>> strategy = rand_strategy()
        >>> print(strategy)
        <abc: utility-dependent (0.5488135039273248, 0.7151893663724195), \
acb: utility-dependent (0.6027633760716439, 0.5448831829968969), \
bac: utility-dependent (0.4236547993389047, 0.6458941130666561), \
bca: utility-dependent (0.4375872112626925, 0.8917730007820798), \
cab: utility-dependent (0.9636627605010293, 0.3834415188257777), \
cba: utility-dependent (0.7917250380826646, 0.5288949197529045)>

    Specify a profile:

        >>> from poisson_approval import ProfileHistogram
        >>> profile = ProfileHistogram({'abc': 0.75, 'bac': 0.25}, {'abc': [1], 'bac': [1]})
        >>> rand_strategy = RandStrategyThresholdUniform(profile=profile)
        >>> strategy = rand_strategy()
        >>> print(strategy)
        <abc: utility-dependent (0.5680445610939323), bac: utility-dependent (0.925596638292661)> ==> a

    Specify some fixed strategies:

        >>> rand_strategy = RandStrategyThresholdUniform(
        ...     d_ranking_fixed_strategy={'abc': 1, 'acb': 1, 'bac': 1, 'bca': (0.5, 0.5)})
        >>> print(rand_strategy())
        <abc: a, acb: a, bac: b, bca: utility-dependent (0.5, 0.5), \
cab: utility-dependent (0.07103605819788694, 0.08712929970154071), \
cba: utility-dependent (0.02021839744032572, 0.832619845547938)>

    Give the ratio of optimistic voters explicitly:

        >>> rand_strategy = RandStrategyThresholdUniform(
        ...     d_ranking_fixed_strategy={'abc': 1, 'acb': 1, 'bac': 1, 'bca': 1},
        ...     ratio_optimistic=0.42)
        >>> print(rand_strategy())
        <abc: a, acb: a, bac: b, bca: b, cab: utility-dependent (0.7781567509498505, 0.42), \
cba: utility-dependent (0.8700121482468192, 0.42)>
    """

    def __init__(self, profile=None, voting_rule=None, d_ranking_fixed_strategy=None, ratio_optimistic=None):
        # Default parameters
        if d_ranking_fixed_strategy is None:
            d_ranking_fixed_strategy = dict()
        # Parameters
        self.profile = profile
        # noinspection PyProtectedMember
        self.voting_rule = Strategy._get_voting_rule_(profile, voting_rule)
        self.d_ranking_fixed_strategy = d_ranking_fixed_strategy
        self.ratio_optimistic = ratio_optimistic

        # Computed variables
        self.rankings_to_decide = RANKINGS if self.profile is None else self.profile.support_in_rankings
        self.rankings_to_decide = sorted(self.rankings_to_decide - self.d_ranking_fixed_strategy.keys())
        self.n_rankings_to_decide = len(self.rankings_to_decide)

    def __call__(self):
        if self.ratio_optimistic is None:
            if self.profile is not None and self.profile.is_continuous:
                d = {ranking: np.random.rand() for ranking in self.rankings_to_decide}
            else:
                d = {ranking: (np.random.rand(), np.random.rand()) for ranking in self.rankings_to_decide}
        else:
            d = {ranking: (np.random.rand(), self.ratio_optimistic) for ranking in self.rankings_to_decide}
        d.update(self.d_ranking_fixed_strategy)
        return StrategyThreshold(d, profile=self.profile, voting_rule=self.voting_rule)



from poisson_approval.constants.basic_constants import *
from poisson_approval.random_factories.RandSimplexGridUniform import RandSimplexGridUniform
from poisson_approval.tau_vector.TauVector import TauVector
from poisson_approval.utils.Util import initialize_random_seeds
from poisson_approval.utils.UtilBallots import allowed_ballots


class RandTauVectorGridUniform(RandSimplexGridUniform):
    """A random factory of tau-vectors (:class:`TauVector`), uniform on a grid.

    Parameters
    ----------
    denominator : int
        The grain of the grid.
    ballots : iterable, optional
        These ballots (e.g. ``'a'``, ``'ab'``) will have a variable share. Default: all allowed ballots.
    d_ballot_fixed_share : dict, optional
        A dictionary. For each entry ``ballot: fixed_share``, this ballot will have at least this fixed share. The total
        must be lower or equal to 1.
    kwargs
        These additional arguments will be passed directly to :class:`TauVector`.

    Examples
    --------
    Basic usage:

        >>> initialize_random_seeds()
        >>> rand_tau = RandTauVectorGridUniform(denominator=7)
        >>> tau = rand_tau()
        >>> print(tau)
        <a: 2/7, ab: 1/7, ac: 1/7, b: 3/7> ==> a, b

    If the voting rule is not approval, only the relevant ballots are used:

        >>> initialize_random_seeds()
        >>> rand_tau = RandTauVectorGridUniform(denominator=7, voting_rule=PLURALITY)
        >>> tau = rand_tau()
        >>> print(tau)
        <a: 2/7, b: 4/7, c: 1/7> ==> b (Plurality)

    Using the optional parameters:

        >>> from fractions import Fraction
        >>> rand_tau = RandTauVectorGridUniform(
        ...     denominator=5,
        ...     ballots=['b', 'bc'], d_ballot_fixed_share={'a': Fraction(2, 7)},
        ...     symbolic=True)
        >>> tau = rand_tau()
        >>> print(tau)
        <a: 2/7, b: 1/7, bc: 4/7> ==> b

    For more examples, cf. :class:`RandSimplexGridUniform`.
    """

    def __init__(self, denominator, ballots=None, d_ballot_fixed_share=None, **kwargs):
        if ballots is None:
            try:
                ballots = allowed_ballots(kwargs['voting_rule'])
            except KeyError:
                ballots = allowed_ballots()
        super().__init__(cls=TauVector, denominator=denominator, keys=ballots,
                         d_key_fixed_share=d_ballot_fixed_share, **kwargs)



from fractions import Fraction
import numpy as np
from poisson_approval.meta_analysis.ternary_plots import ternary_figure
from poisson_approval.profiles.ProfileNoisyDiscrete import ProfileNoisyDiscrete
from poisson_approval.utils.Util import candidates_to_probabilities, one_over_log_t_plus_one, d_candidate_value_to_array


class SimplexToProfile:
    """Map a point of the simplex to a profile (for ternary plots).

    Parameters
    ----------
    cls : class
        :class:`ProfileDiscrete`, :class:`ProfileHistogram`, :class:`ProfileNoisyDiscrete`, :class:`ProfileOrdinal` or
        :class:`ProfileTwelve`.
    right_type,top_type,left_type : object
        A type that is suitable for the kind of profile used.
    d_type_fixed_share : dict
        Key: a type that is suitable for the kind of profile used. Value: a share of voters.
    kwargs
        Other keyword arguments are passed to the profile. E.g. ``voting_rule``, ``ratio_fanatic``, etc.

    Notes
    -----
    A `SimplexToProfile` object is a callable. When called, it takes three parameters `right`, `top`, `left`, numbers
    that represent a point in the simplex. It outputs the profile defined by:

    * The class is given by `cls`,
    * According to `d_type_fixed_share`, some types are assigned fixed shares of voters.
    * The other voters, are distributed between `left_type`, `right_type` and `top_type`, in respective
      proportions that are given by the input tuple `(right, top, left)`.

    Examples
    --------
    Typical usage:

        >>> simplex_to_profile = SimplexToProfile(
        ...     ProfileNoisyDiscrete,
        ...     left_type=('abc', 0.5, 0.01), right_type=('bac', 0.5, 0.01), top_type='c>a~b',
        ...     d_type_fixed_share={('abc', 0.1, 0.01): Fraction(1, 10), ('abc', 0.9, 0.01): Fraction(2, 10)})
        >>> profile = simplex_to_profile(left=Fraction(11, 80), top=Fraction(52, 80), right=Fraction(17, 80))
        >>> print(profile)
        <abc 0.1 ± 0.01: 1/10, abc 0.5 ± 0.01: 77/800, abc 0.9 ± 0.01: 1/5, bac 0.5 ± 0.01: 119/800, c>a~b: 91/200> \
(Condorcet winner: a)

    The types with variable share and fixed share may overlap:

        >>> simplex_to_profile = SimplexToProfile(
        ...     ProfileNoisyDiscrete,
        ...     left_type=('abc', 0.5, 0.01), right_type=('bac', 0.5, 0.01), top_type='c>a~b',
        ...     d_type_fixed_share={('abc', 0.5, 0.01): Fraction(93, 100)})
        >>> profile = simplex_to_profile(left=Fraction(4, 7), top=Fraction(2, 7), right=Fraction(1, 7))
        >>> print(profile)
        <abc 0.5 ± 0.01: 97/100, bac 0.5 ± 0.01: 1/100, c>a~b: 1/50> (Condorcet winner: a)
    """

    def __init__(self, cls, right_type, top_type, left_type, d_type_fixed_share=None, **kwargs):
        if d_type_fixed_share is None:
            d_type_fixed_share = dict()
        self.cls = cls
        self.right_type, self.top_type, self.left_type = right_type, top_type, left_type
        self.d_type_fixed_share = d_type_fixed_share
        self.total_fixed_share = sum(self.d_type_fixed_share.values())
        self.variable_share = 1 - sum(self.d_type_fixed_share.values())
        self.kwargs = kwargs
        self.order_r, self.label_r = cls.order_and_label(right_type)
        self.order_t, self.label_t = cls.order_and_label(top_type)
        self.order_l, self.label_l = cls.order_and_label(left_type)
        self.d_order_fixed_share = dict()
        for t, share in d_type_fixed_share.items():
            order, _ = cls.order_and_label(t)
            self.d_order_fixed_share[order] = self.d_order_fixed_share.get(order, 0) + share

    def __call__(self, right, top, left):
        d_type_share = dict()
        d_type_share.update(self.d_type_fixed_share)
        d_type_share[self.right_type] = d_type_share.get(self.right_type, 0) + right * self.variable_share
        d_type_share[self.top_type] = d_type_share.get(self.top_type, 0) + top * self.variable_share
        d_type_share[self.left_type] = d_type_share.get(self.left_type, 0) + left * self.variable_share
        return self.cls(d_type_share, **self.kwargs)


def ternary_plot_n_equilibria(simplex_to_profile, scale, title='Number of equilibria',
                              meth='analyzed_strategies_ordinal', **kwargs):
    """Shortcut: ternary plot for the number of equilibria.

    Parameters
    ----------
    simplex_to_profile : SimplexToProfile
        This is responsible for generating the profiles.
    scale : Number
        Scale of the plot (resolution).
    title : str
        Title of the plot.
    meth : str
        The name of the :class:`AnalyzedStrategies` property used to count the equilibria. Cf. :class:`Profile`.
    kwargs
        Other keyword arguments are passed to the function :meth:`TernaryAxesSubplotPoisson.heatmap_intensity`.

    Examples
    --------
        >>> simplex_to_profile = SimplexToProfile(
        ...     ProfileNoisyDiscrete,
        ...     left_type=('abc', 0.5, 0.01), right_type=('bac', 0.5, 0.01), top_type=('cab', 0.5, 0.01))
        >>> figure, tax = ternary_plot_n_equilibria(simplex_to_profile, scale=10)
    """
    def n_equilibria(right, top, left):
        profile = simplex_to_profile(right, top, left)
        return len(getattr(profile, meth).equilibria)
    figure, tax = ternary_figure(scale=scale)
    tax.heatmap_intensity(n_equilibria,
                          right_label=simplex_to_profile.label_r,
                          top_label=simplex_to_profile.label_t,
                          left_label=simplex_to_profile.label_l,
                          **kwargs)
    tax.annotate_condorcet(right_order=simplex_to_profile.order_r,
                           top_order=simplex_to_profile.order_t,
                           left_order=simplex_to_profile.order_l,
                           d_order_fixed_share=simplex_to_profile.d_order_fixed_share)
    tax.set_title_padded(title)
    return figure, tax


def ternary_plot_winners_at_equilibrium(simplex_to_profile, scale, title='Winners at equilibrium',
                                        legend_title='Winners', meth='analyzed_strategies_ordinal',
                                        file_save_data=None,
                                        **kwargs):
    """Shortcut: ternary plot for the winners at equilibrium.

    Parameters
    ----------
    simplex_to_profile : SimplexToProfile
        This is responsible for generating the profiles.
    scale : Number
        Scale of the plot (resolution).
    title : str
        Title of the plot.
    legend_title : str
        Title of the legend of the plot.
    meth : str
        The name of the :class:`AnalyzedStrategies` property used to study the equilibria. Cf. :class:`Profile`.
    file_save_data : str
        File where the computed data will be saved (using ``pickle``).
    kwargs
        Other keyword arguments are passed to the function :meth:`TernaryAxesSubplotPoisson.heatmap_candidates`.

    Examples
    --------
        >>> simplex_to_profile = SimplexToProfile(
        ...     ProfileNoisyDiscrete,
        ...     left_type=('abc', 0.5, 0.01), right_type=('bac', 0.5, 0.01), top_type=('cab', 0.5, 0.01))
        >>> figure, tax = ternary_plot_winners_at_equilibrium(simplex_to_profile, scale=10)
    """

    def winners_at_equilibrium(right, top, left):
        profile = simplex_to_profile(right, top, left)
        return candidates_to_probabilities(getattr(profile, meth).winners_at_equilibrium)
    figure, tax = ternary_figure(scale=scale)
    tax.heatmap_candidates(winners_at_equilibrium,
                           right_label=simplex_to_profile.label_r,
                           top_label=simplex_to_profile.label_t,
                           left_label=simplex_to_profile.label_l,
                           legend_style='color_patches',
                           legend_title=legend_title,
                           file_save_data=file_save_data,
                           **kwargs)
    tax.annotate_condorcet(right_order=simplex_to_profile.order_r,
                           top_order=simplex_to_profile.order_t,
                           left_order=simplex_to_profile.order_l,
                           d_order_fixed_share=simplex_to_profile.d_order_fixed_share)
    tax.set_title_padded(title)
    return figure, tax


def ternary_plot_winning_frequencies(simplex_to_profile, scale,
                                     n_max_episodes, init='sincere', samples_per_point=1,
                                     perception_update_ratio=one_over_log_t_plus_one,
                                     ballot_update_ratio=one_over_log_t_plus_one,
                                     winning_frequency_update_ratio=one_over_log_t_plus_one,
                                     title='Winning frequencies', legend_title='Winners',
                                     meth='fictitious_play', file_save_data=None,
                                     **kwargs):
    """Shortcut: ternary plot for the winning frequencies in fictitious play / iterated voting.

    Parameters
    ----------
    simplex_to_profile : SimplexToProfile
        This is responsible for generating the profiles.
    scale : Number
        Scale of the plot (resolution).
    n_max_episodes : int
        Maximum number of episodes for the fictitious play / iterated voting.
    init : Strategy or TauVector or str
        Cf. :meth:`~poisson_approval.ProfileCardinal.fictitious_play` or
        :meth:`~poisson_approval.ProfileCardinal.iterated_voting`.
    samples_per_point : int
        How many trials are made for each point drawn. Useful only when initialization is random.
    perception_update_ratio,ballot_update_ratio,winning_frequency_update_ratio : callable or Number
        Cf. :meth:`~poisson_approval.ProfileCardinal.fictitious_play` or
        :meth:`~poisson_approval.ProfileCardinal.iterated_voting`.
    title : str
        Title of the plot.
    legend_title : str
        Title of the legend of the plot.
    meth : str
        The name of the method (``'fictitious_play'`` or ``'iterated_voting'``).
    file_save_data : str
        File where the computed data will be saved (using ``pickle``).
    kwargs
        Other keyword arguments are passed to the function :meth:`TernaryAxesSubplotPoisson.heatmap_candidates`.

    Examples
    --------
        >>> simplex_to_profile = SimplexToProfile(
        ...     ProfileNoisyDiscrete,
        ...     left_type=('abc', 0.5, 0.01), right_type=('bac', 0.5, 0.01), top_type=('cab', 0.5, 0.01))
        >>> figure, tax = ternary_plot_winning_frequencies(simplex_to_profile, scale=10, n_max_episodes=10)
    """
    def winning_frequencies(right, top, left):
        profile = simplex_to_profile(right, top, left)
        a_candidate_value = np.zeros(3)
        for _ in range(samples_per_point):
            results = getattr(profile, meth)(init=init, n_max_episodes=n_max_episodes,
                                             perception_update_ratio=perception_update_ratio,
                                             ballot_update_ratio=ballot_update_ratio,
                                             winning_frequency_update_ratio=winning_frequency_update_ratio)
            a_candidate_value = a_candidate_value + d_candidate_value_to_array(results['d_candidate_winning_frequency'])
        return a_candidate_value / samples_per_point

    figure, tax = ternary_figure(scale=scale)
    tax.heatmap_candidates(winning_frequencies,
                           right_label=simplex_to_profile.label_r,
                           top_label=simplex_to_profile.label_t,
                           left_label=simplex_to_profile.label_l,
                           legend_style='palette',
                           legend_title=legend_title,
                           file_save_data=file_save_data,
                           **kwargs)
    tax.annotate_condorcet(right_order=simplex_to_profile.order_r,
                           top_order=simplex_to_profile.order_t,
                           left_order=simplex_to_profile.order_l,
                           d_order_fixed_share=simplex_to_profile.d_order_fixed_share)
    tax.set_title_padded(title)
    return figure, tax


def ternary_plot_convergence(simplex_to_profile, scale,
                             n_max_episodes, init='sincere', samples_per_point=1,
                             perception_update_ratio=one_over_log_t_plus_one,
                             ballot_update_ratio=one_over_log_t_plus_one,
                             title='Convergence frequency',
                             meth='fictitious_play', **kwargs):
    """Shortcut: ternary plot for the convergence frequency in fictitious play / iterated voting.

    Convergence frequency: out of `samples_per_points` trials, in which proportion of the cases did fictitious play or
    iterated voting converge within `n_max_episodes` iterations?

    Parameters
    ----------
    simplex_to_profile : SimplexToProfile
        This is responsible for generating the profiles.
    scale : Number
        Scale of the plot (resolution).
    n_max_episodes : int
        Maximum number of episodes for the fictitious play / iterated voting.
    init : Strategy or TauVector or str
        Cf. :meth:`~poisson_approval.ProfileCardinal.fictitious_play` or
        :meth:`~poisson_approval.ProfileCardinal.iterated_voting`.
    samples_per_point : int
        How many trials are made for each point drawn. Useful only when initialization is random.
    perception_update_ratio,ballot_update_ratio : callable or Number
        Cf. :meth:`~poisson_approval.ProfileCardinal.fictitious_play` or
        :meth:`~poisson_approval.ProfileCardinal.iterated_voting`.
    title : str
        Title of the plot.
    meth : str
        The name of the method (``'fictitious_play'`` or ``'iterated_voting'``).
    kwargs
        Other keyword arguments are passed to the function :meth:`TernaryAxesSubplotPoisson.heatmap_intensity`.

    Examples
    --------
        >>> simplex_to_profile = SimplexToProfile(
        ...     ProfileNoisyDiscrete,
        ...     left_type=('abc', 0.5, 0.01), right_type=('bac', 0.5, 0.01), top_type=('cab', 0.5, 0.01))
        >>> figure, tax = ternary_plot_convergence(simplex_to_profile, scale=10, n_max_episodes=10)
    """
    def convergence_frequency(right, top, left):
        profile = simplex_to_profile(right, top, left)
        n_convergences = 0
        for _ in range(samples_per_point):
            results = getattr(profile, meth)(init=init, n_max_episodes=n_max_episodes,
                                             perception_update_ratio=perception_update_ratio,
                                             ballot_update_ratio=ballot_update_ratio)
            if results['converges']:
                n_convergences += 1
        return n_convergences / samples_per_point

    figure, tax = ternary_figure(scale=scale)
    tax.heatmap_intensity(convergence_frequency,
                          right_label=simplex_to_profile.label_r,
                          top_label=simplex_to_profile.label_t,
                          left_label=simplex_to_profile.label_l,
                          vmin=0., vmax=1.,
                          **kwargs)
    tax.annotate_condorcet(right_order=simplex_to_profile.order_r,
                           top_order=simplex_to_profile.order_t,
                           left_order=simplex_to_profile.order_l,
                           d_order_fixed_share=simplex_to_profile.d_order_fixed_share)
    tax.set_title_padded(title)
    return figure, tax



import numpy as np
from matplotlib import pyplot as plt
from poisson_approval.utils.UtilPlot import plt_cdf
from poisson_approval.random_factories.RandProfileHistogramUniform import RandProfileHistogramUniform
from poisson_approval.meta_analysis.monte_carlo_fictitious_play import monte_carlo_fictitious_play, \
    MCS_UTILITY_THRESHOLDS


def plot_utility_thresholds(results, voting_rule=None, **kwargs):
    """
    Plot the distribution (CDF) of the utility threshold.

    Parameters
    ----------
    results : dict
        Results of :func:`~poisson_approval.meta_analysis.monte_carlo_fictitious_play.monte_carlo_fictitious_play`,
        with at least the setting
        :const:`~poisson_approval.meta_analysis.monte_carlo_fictitious_play.MCS_UTILITY_THRESHOLDS`.
    voting_rule : str
        The voting rule (or None if this parameter was not given to
        :func:`~poisson_approval.meta_analysis.monte_carlo_fictitious_play.monte_carlo_fictitious_play`).
    kwargs
        Other keyword arguments are passed to the function ``step`` of `matplotlib`.

    Examples
    --------
        >>> meta_results = monte_carlo_fictitious_play(
        ...     factory=RandProfileHistogramUniform(n_bins=1),
        ...     n_samples=1,
        ...     n_max_episodes=10,
        ...     monte_carlo_settings=[MCS_UTILITY_THRESHOLDS],
        ... )
        >>> plot_utility_thresholds(meta_results)
    """
    if voting_rule is None:
        voting_rule = ''
    n_samples = results[voting_rule]['n_samples']
    utility_thresholds = np.array(results[voting_rule]['utility_thresholds']).flatten()
    weights = np.array(results[voting_rule]['weights_rankings']).flatten() / n_samples
    fig, ax = plt.subplots(figsize=(8, 4))
    plt_cdf(utility_thresholds, weights, n_samples, label='Utility', **kwargs)
    ax.grid(True)
    # ax.set_title('Utility threshold')
    ax.set_xlabel('Utility threshold')
    ax.set_ylabel('Cumulative likelihood of occurrence')
    plt.xlim(0, 1)
    plt.xticks(np.arange(0, 1.1, 0.1))



import pickle
import ternary
from math import floor, ceil
from fractions import Fraction
from collections import Counter, namedtuple
from scipy.spatial import distance
from ternary.helpers import simplex_iterator
from matplotlib import pyplot as plt
from matplotlib.patches import Patch
from poisson_approval.meta_analysis.ternary_condorcet import draw_condorcet_zones
from poisson_approval.meta_analysis.colors import *


Point = namedtuple('Point', ['right', 'top', 'left'])


def _generate_heatmap_data(f, scale):
    """Generate RGBA data for a ``simplex to 3D'' heatmap plot.

    Parameters
    ----------
    f : callable
        The function to plot. Input: coordinates `right`, `top`, `left` in the simplex, i.e. that sum to 1. Output: a
        list of 3 numbers between 0 and 1.
    scale
        The scale of the ternary plot.

    Returns
    -------
    d_scaled_point_color : dict
        Key: a point of the integer simplex defined by `scale`. Value: the RGBA code of the color representing the
        output of `f`.
    d_point_values : dict
        Key: a point of the simplex, with coordinates in (0, 1). Value: the values of function f(right, top, left).

    Examples
    --------
        >>> def f(right, top, left):
        ...     return [right, 0, 0]
        >>> d_scaled_point_color, d_point_values = _generate_heatmap_data(f, scale=2)
        >>> d_scaled_point_color
        {(0, 0, 2): (0.5, 0.5, 0.5, 1.0), (0, 1, 1): (0.5, 0.5, 0.5, 1.0), (0, 2, 0): (0.5, 0.5, 0.5, 1.0), \
(1, 0, 1): (0.75, 0.5, 0.5, 1.0), (1, 1, 0): (0.75, 0.5, 0.5, 1.0), (2, 0, 0): (1.0, 0.5, 0.5, 1.0)}
        >>> d_point_values  # doctest: +ELLIPSIS
        {Point(right=Fraction(0, 1), top=Fraction(0, 1), left=Fraction(1, 1)): [Fraction(0, 1), 0, 0], ...
    """
    d_point_values = dict()
    d_scaled_point_color = dict()
    for (right, top, left) in simplex_iterator(scale):
        scaled_point = (right, top, left)
        point = Point(Fraction(right, scale), Fraction(top, scale), Fraction(left, scale))
        values = f(*point)
        d_point_values[point] = values
        color = abc_to_rgb(values)
        d_scaled_point_color[scaled_point] = (float(color[0]), float(color[1]), float(color[2]), 1.)
    return d_scaled_point_color, d_point_values


def ternary_figure(size_inches='auto', scale=None, boundary_width=1.0, **kwargs):
    """Create a ternary plot (adaptation of ``figure`` from the package `python-ternary`).

    Parameters
    ----------
    size_inches : tuple or str
        The horizontal and vertical sizes of the figure, in inches. If ``'auto'``, we will try to do our best.
    scale : Number
        The scale of the figure. The higher it is, the higher the resolution of the heatmap for example.
    boundary_width
        Width of the line representing the boundary of the triangle.
    kwargs
        Other keyword arguments are passed to the function ``figure`` of the package `python-ternary`.

    Returns
    -------
    figure : matplotlib.figure.Figure
    ternary_ax : :class:`TernaryAxesSubplotPoisson`

    Examples
    --------
    Grid and lines:

        >>> figure, tax = ternary_figure()
        >>> tax.gridlines_simplex(multiple=0.1)
        >>> tax.horizontal_line_simplex(0.1)
        >>> tax.left_parallel_line_simplex(0.2)
        >>> tax.right_parallel_line_simplex(0.3)
        >>> tax.line_simplex((1, 0, 0), (0.2, 0.5, 0.3))

    Titles and annotations:

        >>> figure, tax = ternary_figure()
        >>> tax.set_title_padded('The title of the figure')
        >>> tax.right_corner_label('The right label')
        >>> tax.top_corner_label('The top label')
        >>> tax.left_corner_label('The left label')
        >>> tax.annotate_simplex('An annotation', (0.33, 0.33, 0.33))
    """
    ternary_ax = TernaryAxesSubplotPoisson(scale=scale, size_inches=size_inches, **kwargs)
    figure = ternary_ax.get_figure()
    if size_inches == 'auto':
        figure.set_size_inches(5, 5)
    else:
        figure.set_size_inches(*size_inches)
    plt.axis('off')
    ternary_ax.boundary(linewidth=boundary_width)
    return figure, ternary_ax


class TernaryAxesSubplotPoisson(ternary.TernaryAxesSubplot):
    """Subclass of ``TernaryAxesSubplot``, defined in the package `python-ternary`.

    This class implements some additional methods for Poisson Approval. For some examples, cf. the corresponding
    tutorial.
    """

    def __init__(self, scale=None, size_inches='auto', **kwargs):
        self.size_inches = size_inches
        self.d_point_values_ = None  # Used for candidate maps
        super().__init__(scale=scale, **kwargs)

    def _scaled_number(self, x):
        """Scale a number.

        Parameters
        ----------
        x : Number
            A coordinate in [0, 1].

        Returns
        -------
        Number
            A coordinate in [0, ``self.scale``].
        """
        return x * self.get_scale()

    def _scaled_position(self, p):
        """Scale a position.

        Parameters
        ----------
        p : tuple
            Three coordinates in [0, 1] (typically, that sum to 1).

        Returns
        -------
        Tuple
            Three coordinates in [0, ``self.scale``].
        """
        return [self._scaled_number(x) for x in p]

    def set_title_padded(self, title, **kwargs):
        """Adaptation of ``set_title``.

        Adjust the position of the title to avoid collision with the label of the top corner.

        Notes
        -----
        For the parameters, cf. ``set_title`` in `python-ternary`.
        """
        size_inches = plt.gcf().get_size_inches()
        pad = 6 * size_inches[1]
        pad += kwargs.pop('pad', 0)
        self.set_title(title, pad=pad, **kwargs)

    def annotate_simplex(self, text, position, horizontalalignment='center', verticalalignment='center', **kwargs):
        """Adaptation of ``annotate``.

        * The input position is in the simplex (instead of the scaled simplex), i.e. coordinates are in [0, 1].
        * The default horizontal and vertical alignments are centered.

        Notes
        -----
        For the parameters, cf. ``annotate`` in `python-ternary`.
        """
        self.annotate(text, position=self._scaled_position(position),
                      horizontalalignment=horizontalalignment, verticalalignment=verticalalignment, **kwargs)

    def gridlines_simplex(self, multiple=None, horizontal_kwargs=None, left_kwargs=None,
                          right_kwargs=None, **kwargs):
        """Adaptation of ``gridlines``.

        * The argument `multiple` is a grid step in [0, 1] (instead of [0, scale]).

        Notes
        -----
        For the parameters, cf. ``gridlines`` in `python-ternary`.
        """
        self.gridlines(multiple=self._scaled_number(multiple), horizontal_kwargs=horizontal_kwargs,
                       left_kwargs=left_kwargs, right_kwargs=right_kwargs, **kwargs)

    def line_simplex(self, p1, p2, color='black', **kwargs):
        """Adaptation of ``line``.

        * The input positions are in the simplex (instead of the scaled simplex), i.e. coordinates are in [0, 1].

        Notes
        -----
        For the parameters, cf. ``line`` in `python-ternary`.
        """
        self.line(self._scaled_position(p1), self._scaled_position(p2), color=color, **kwargs)

    def horizontal_line_simplex(self, i, color='black', **kwargs):
        """Adaptation of ``horizontal_line``.

        * The argument `i` is in [0, 1] (instead of [0, scale]).

        Notes
        -----
        For the parameters, cf. ``horizontal_line`` in `python-ternary`.
        """
        self.horizontal_line(self._scaled_number(i), color=color, **kwargs)

    def left_parallel_line_simplex(self, i, color='black', **kwargs):
        """Adaptation of ``left_parallel_line``.

        * The argument `i` is in [0, 1] (instead of [0, scale]).

        Notes
        -----
        For the parameters, cf. ``left_parallel_line`` in `python-ternary`.
        """
        self.left_parallel_line(self._scaled_number(i), color=color, **kwargs)

    def right_parallel_line_simplex(self, i, color='black', **kwargs):
        """Adaptation of ``right_parallel_line``.

        * The argument `i` is in [0, 1] (instead of [0, scale]).

        Notes
        -----
        For the parameters, cf. ``right_parallel_line`` in `python-ternary`.
        """
        self.right_parallel_line(self._scaled_number(i), color=color, **kwargs)

    def heatmap_intensity(self, func, right_label, top_label, left_label,
                          style='hexagonal', cmap='plasma', **kwargs):
        """Adaptation of ``heatmapf``.

        Parameters
        ----------
        func : callable
            The function to plot. Input: coordinates `right`, `top`, `left` in the simplex, i.e. that sum to 1.
            Output: a number.
        right_label : str
            Label of the right corner.
        top_label : str
            Label of the top corner.
        left_label : str
            Label of the left corner.
        style : str
            Contrarily to default settings in `python-ternary`, the default is ``'hexagonal'``.
        cmap : str
            Colormap. Contrarily to default settings in `python-ternary`, the default is ``'plasma'``.
        kwargs
            All other keywords arguments are passed to method ``heatmapf`` of `python-ternary`.

        Examples
        --------
            >>> def f(right, top, left):
            ...     return (right**2 + top) / (left + 1)
            >>> figure, tax = ternary_figure(scale=10)
            >>> tax.heatmap_intensity(f,
            ...                       left_label='left',
            ...                       right_label='right',
            ...                       top_label='top')
            >>> tax.set_title_padded('An intensity heat map')
        """
        default_pad = 0.15
        if 'cb_kwargs' not in kwargs.keys():
            kwargs['cb_kwargs'] = {'pad': default_pad}
        elif 'pad' not in kwargs['cb_kwargs'].keys():
            kwargs['cb_kwargs']['pad'] = default_pad
        self.heatmapf(lambda p: func(*p), style=style, cmap=cmap, **kwargs)
        self.right_corner_label(right_label)
        self.top_corner_label(top_label)
        self.left_corner_label(left_label)
        if self.size_inches == 'auto':
            plt.gcf().set_size_inches(7, 5)

    def heatmap_candidates(self, func, right_label, top_label, left_label, legend_title='',
                           legend_style='palette', style='hexagonal', colorbar=False, file_save_data = None,
                           **kwargs):
        """Heatmap of a function from the simplex to 3D vectors.

        Parameters
        ----------
        func: callable
            The function to plot. Input: coordinates `right`, `top`, `left` in the simplex, i.e. that sum to 1.
            Output: a list of 3 numbers between 0 and 1.
        right_label : str
            Label of the right corner.
        top_label : str
            Label of the top corner.
        left_label : str
            Label of the left corner.
        legend_title : str
            Title of the legend.
        legend_style : str
            The style of the legend. The two available options are ``'palette'`` and ``'color_patches'``.
            Cf. :meth:`legend_palette` and :meth:`legend_color_patches`.
        style
            Contrarily to default settings in `python-ternary`, the default is ``'hexagonal'``.
        colorbar
            Contrarily to default settings in `python-ternary`, the default is False.
        file_save_data : str
            File where the computed data will be saved (using ``pickle``).
        kwargs
            All other keywords arguments are passed to method ``heatmap`` of `python-ternary`.

        Examples
        --------
            >>> def g(right, top, left):
            ...     a = top**.5
            ...     b = left**2.
            ...     c = 1 - a - b
            ...     return [a, b, c]
            >>> figure, tax = ternary_figure(scale=5)
            >>> tax.heatmap_candidates(g,
            ...                        left_label='left',
            ...                        right_label='right',
            ...                        top_label='top',
            ...                        legend_title='Candidates',
            ...                        legend_style='palette')
            >>> tax.set_title_padded('A candidate heat map')

        To retrieve the computed value that is closest to a given point:

            >>> tax.f_point_values_(right=0.5, top=0.3, left=0.2)
            [0.4472135954999579, 0.04000000000000001, 0.5127864045000421]
        """
        d_scaled_point_color, self.d_point_values_ = _generate_heatmap_data(func, self.get_scale())
        if file_save_data is not None:
            with open(file_save_data, "wb") as f:
                pickle.dump([d_scaled_point_color, self.d_point_values_], f)
        self.heatmap(d_scaled_point_color, style=style, colorbar=colorbar, use_rgba=True, **kwargs)
        self.right_corner_label(right_label)
        self.top_corner_label(top_label)
        self.left_corner_label(left_label)
        if self.size_inches == 'auto':
            plt.gcf().set_size_inches(5, 5)
        if legend_style == 'palette':
            self.legend_palette(title=legend_title)
        else:
            self.legend_color_patches(title=legend_title, data=d_scaled_point_color)

    def f_point_values_(self, right, top, left):
        """Data of a candidate heatmap.

        This function can be used only after a candidate heatmap has been drawn. When given a point of the simplex,
        it considers the closest point where `f` was computed and returns the corresponding value. It does not
        recompute `f`, and in particular the result may not be the same as a direct application of `f` to the point
        given as input.

        Parameters
        ----------
        right : Number
        top : Number
        left : Number

        Returns
        -------
        list
            List of size 3. The value associated respectively to each candidate `a`, `b`, `c`.
        """
        if self.d_point_values_ is None:
            raise ValueError("No candidate heatmap has been defined")
        scale = self.get_scale()
        input_point = Point(right=right, top=top, left=left)
        possible_approx_points = [
            Point(right=ri, top=to, left=le)
            for ri in [Fraction(floor(right * scale), scale), Fraction(ceil(right * scale), scale)]
            for to in [Fraction(floor(top * scale), scale), Fraction(ceil(top * scale), scale)]
            for le in [Fraction(floor(left * scale), scale), Fraction(ceil(left * scale), scale)]
            if ri + to + le == 1
        ]
        best_approx_point = min(possible_approx_points, key=lambda p: distance.euclidean(input_point, p))
        return self.d_point_values_[best_approx_point]

    def annotate_condorcet(self, right_order, top_order, left_order, d_order_fixed_share=None):
        """Annotate who is the Condorcet winner depending on the region.

        We consider a simplex where:

        * Fixed shares of voters have preference orders given by `d_order_fixed_share`.
        * The remaining voters are split between `right_order`, `top_order` and `left_order` in proportions that are
          given by the point in the simplex.

        This method annotates the regions according to which candidate is the Condorcet winner, and indicates where no
        one is the Condorcet winner.

        If there are weak orders and/or fixed shares of voters, the method may not work on all distributions because it
        relies on an external package called `shapely`. If there are rankings only and no fixed shares of voters, it is
        supposed to work on all distributions.

        Parameters
        ----------
        right_order : str
            The order whose share is maximal at the right corner.
        top_order : str
            The order whose share is maximal at the top corner.
        left_order : str
            The order whose share is maximal at the left corner.
        d_order_fixed_share : dict, optional
            Key: order. Value: a fixed share of voters in [0, 1].
        """
        try:
            draw_condorcet_zones(self, right_order, top_order, left_order, d_order_fixed_share)
        except NameError:  # pragma: no cover
            self._annotate_condorcet_old(right_order, top_order, left_order, d_order_fixed_share)

    def _annotate_condorcet_old(self, right_ranking, top_ranking, left_ranking, d_order_fixed_share=None):
        """Old version, for rankings only"""
        if d_order_fixed_share:
            raise NotImplementedError
        count_candidate_tops = dict(Counter([
            ranking[0] for ranking in [right_ranking, top_ranking, left_ranking]]))
        if len(count_candidate_tops) == 1:
            # All rankings have the same top: 'a..', 'a..', 'a..'
            candidate = right_ranking[0]
            self.annotate_simplex('%s is CW' % candidate, (0.33, 0.33, 0.33))
        elif len(count_candidate_tops) == 2:
            # Two different tops: 'a..', 'a..', 'b..'
            weak_candidate, strong_candidate = sorted(count_candidate_tops.keys(),
                                                      key=count_candidate_tops.get)
            if right_ranking[0] == weak_candidate:
                self.left_parallel_line_simplex(0.5)
                self.annotate_simplex('%s is CW' % weak_candidate, (0.7, 0.15, 0.15))
                self.annotate_simplex('%s is CW' % strong_candidate, (0.25, 0.375, 0.375))
            elif top_ranking[0] == weak_candidate:
                self.horizontal_line_simplex(0.5)
                self.annotate_simplex('%s is CW' % weak_candidate, (0.15, 0.7, 0.15))
                self.annotate_simplex('%s is CW' % strong_candidate, (0.375, 0.25, 0.375))
            else:  # left_ranking[0] == weak_candidate
                self.right_parallel_line_simplex(0.5)
                self.annotate_simplex('%s is CW' % weak_candidate, (0.15, 0.15, 0.7))
                self.annotate_simplex('%s is CW' % strong_candidate, (0.375, 0.375, 0.25))
        else:
            # Three different tops
            count_candidate_bottoms = dict(Counter([
                ranking[2] for ranking in [right_ranking, top_ranking, left_ranking]]))
            if len(count_candidate_bottoms) == 3:
                # Condorcet cycle: 'abc', 'bca', 'cab'
                self.horizontal_line_simplex(0.5)
                self.left_parallel_line_simplex(0.5)
                self.right_parallel_line_simplex(0.5)
                self.annotate_simplex('%s is CW' % right_ranking[0], (0.7, 0.15, 0.15))
                self.annotate_simplex('%s is CW' % top_ranking[0], (0.15, 0.7, 0.15))
                self.annotate_simplex('%s is CW' % left_ranking[0], (0.15, 0.15, 0.7))
                self.annotate_simplex('no CW', (0.33, 0.33, 0.33))
            else:
                # Divided majority: 'abc', 'bac', 'cab'
                middle_candidate, weak_candidate = sorted(count_candidate_bottoms.keys(),
                                                          key=count_candidate_bottoms.get)
                strong_candidate = next(
                    candidate for candidate in {'a', 'b', 'c'} - {middle_candidate, weak_candidate})
                if right_ranking[0] == strong_candidate:
                    self.horizontal_line_simplex(0.5)
                    self.right_parallel_line_simplex(0.5)
                    self.annotate_simplex('%s is CW' % right_ranking[0], (0.5, 0.25, 0.25))
                    self.annotate_simplex('%s is CW' % top_ranking[0], (0.15, 0.7, 0.15))
                    self.annotate_simplex('%s is CW' % left_ranking[0], (0.15, 0.15, 0.7))
                elif top_ranking[0] == strong_candidate:
                    self.left_parallel_line_simplex(0.5)
                    self.right_parallel_line_simplex(0.5)
                    self.annotate_simplex('%s is CW' % right_ranking[0], (0.7, 0.15, 0.15))
                    self.annotate_simplex('%s is CW' % top_ranking[0], (0.25, 0.5, 0.25))
                    self.annotate_simplex('%s is CW' % left_ranking[0], (0.15, 0.15, 0.7))
                else:  # left_ranking[0] == strong_candidate
                    self.horizontal_line_simplex(0.5)
                    self.left_parallel_line_simplex(0.5)
                    self.annotate_simplex('%s is CW' % right_ranking[0], (0.7, 0.15, 0.15))
                    self.annotate_simplex('%s is CW' % top_ranking[0], (0.15, 0.7, 0.15))
                    self.annotate_simplex('%s is CW' % left_ranking[0], (0.25, 0.25, 0.5))

    def legend_color_patches(self, title, all_mixes=False, data=None):
        """Add a "color patches" legend to the current figure (for candidates heat maps).

        Parameters
        ----------
        title : str
            Title of the legend.
        all_mixes : bool
            If True, then all mixes are indicated in the legend: `a + b`, `a + c`, `b + c` and `a + b + c`.
        data : dict, optional
            Key : anything. Value: a tuple, list, etc giving an RGB or RGBA code. This is used only if `all_mixes`
            is False.

        Notes
        -----
        * The color patches for single candidates `a`, `b`, `c` are in the legend anyway.
        * If `all_mixes` is True, then all mixes are indicated in the legend.
        * If `all_mixes` is False and `data` is given, then if at least a value is close to a perfect mix (like half
          of `a` and half of `b`), then this mix is added to the legend.
        """
        def _edge_color(color):
            return [max(x - .4, 0) for x in color]
        legend_elements = [Patch(facecolor=COLOR_A, edgecolor=_edge_color(COLOR_A), label='$a$'),
                           Patch(facecolor=COLOR_B, edgecolor=_edge_color(COLOR_B), label='$b$'),
                           Patch(facecolor=COLOR_C, edgecolor=_edge_color(COLOR_C), label='$c$')]
        if all_mixes or data is not None:
            def add_optional_patch(color, label, insertion=False):
                if all_mixes or (data is not None
                                 and any(np.all(np.isclose(data_col[0:3], color, atol=1E-1))
                                         for data_col in data.values())):
                    patch = Patch(facecolor=color, edgecolor=_edge_color(color), label=label)
                    if insertion:
                        legend_elements.insert(0, patch)
                    else:
                        legend_elements.append(patch)
            add_optional_patch(COLOR_NO_ONE, 'No one', insertion=True)
            add_optional_patch(COLOR_AB, 'a, b')
            add_optional_patch(COLOR_AC, 'a, c')
            add_optional_patch(COLOR_BC, 'b, c')
            add_optional_patch(COLOR_ABC, 'a, b, c')
        self.legend(title=title, handles=legend_elements)

    @staticmethod
    def legend_palette(title):
        """Add a "palette" legend to the current figure (for candidates heat maps).

        Parameters
        ----------
        title : str
            Title of the legend.

        Notes
        -----
        This is more a hack than a real `matplotlib` legend. Hence the usual commands to choose the position
        of the legend, for example, are not meant to work here.
        """
        legend_ax = plt.gcf().add_axes([.68, .63, .2, .2], facecolor='k')
        legend_ax.imshow(DATA_PALETTE, origin='lower', interpolation='gaussian')
        legend_ax.set_xticks([])
        legend_ax.set_yticks([])
        legend_ax.axis('off')
        legend_ax.annotate('a', (100, 169), horizontalalignment='center', verticalalignment='center')
        legend_ax.annotate('b', (35, 63), horizontalalignment='center', verticalalignment='center')
        legend_ax.annotate('c', (165, 63), horizontalalignment='center', verticalalignment='center')
        legend_ax.set_title(title)



from fractions import Fraction
import numpy as np
from poisson_approval import ProfileHistogram, StrategyThreshold, StrategyOrdinal, EquilibriumStatus, PLURALITY, \
    ANTI_PLURALITY, initialize_random_seeds


def test_normalization():
    profile = ProfileHistogram(d_ranking_share={'abc': 1, 'acb': 1},
                               d_ranking_histogram = {'abc': [1, 1], 'acb': [1]})
    assert profile.abc == 0.5
    assert np.all(profile.d_ranking_histogram['abc'] == [0.5, 0.5])


def test_not_equilibrium():
    profile = ProfileHistogram(d_ranking_share={'abc': 0.5, 'acb': 0.5},
                               d_ranking_histogram = {'abc': [1], 'acb': [1]})
    strategy = StrategyThreshold({'abc': 0, 'acb': 0})
    assert profile.is_equilibrium(strategy) == EquilibriumStatus.NOT_EQUILIBRIUM


def test_no_atom():
    """
        >>> from fractions import Fraction
        >>> profile = ProfileHistogram(
        ...     {'abc': Fraction(1, 10), 'bac': Fraction(6, 10), 'cab': Fraction(3, 10)},
        ...     {'abc': [1], 'bac': [1, 0], 'cab': [Fraction(2, 3), 0, 0, 0, 0, 0, 0, 0, 0, Fraction(1, 3)]})
        >>> profile.have_ranking_with_utility_u(ranking='cab', u=Fraction(1, 100))
        0
    """
    pass


def test_plurality():
    """
        >>> profile = ProfileHistogram(d_ranking_share={'abc': 1, 'acb': 1},
        ...                            d_ranking_histogram = {'abc': [1, 1], 'acb': [1]},
        ...                            voting_rule=PLURALITY)
        >>> profile
        ProfileHistogram({'abc': Fraction(1, 2), 'acb': Fraction(1, 2)}, \
{'abc': array([Fraction(1, 2), Fraction(1, 2)], dtype=object), \
'acb': array([1])}, voting_rule='Plurality')
        >>> print(profile)
        <abc: 1/2 [Fraction(1, 2) Fraction(1, 2)], acb: 1/2 [1]> (Condorcet winner: a) (Plurality)
    """
    pass


def test_anti_plurality():
    """
        >>> profile = ProfileHistogram(d_ranking_share={'abc': 1, 'acb': 1},
        ...                            d_ranking_histogram = {'abc': [1, 1], 'acb': [1]},
        ...                            voting_rule=ANTI_PLURALITY)
        >>> profile
        ProfileHistogram({'abc': Fraction(1, 2), 'acb': Fraction(1, 2)}, \
{'abc': array([Fraction(1, 2), Fraction(1, 2)], dtype=object), 'acb': array([1])}, \
voting_rule='Anti-plurality')
        >>> print(profile)
        <abc: 1/2 [Fraction(1, 2) Fraction(1, 2)], acb: 1/2 [1]> (Condorcet winner: a) (Anti-plurality)
    """
    pass


def test_strategy_weak_order():
    """
        >>> profile = ProfileHistogram(d_ranking_share={'abc': Fraction(1, 7)}, d_ranking_histogram = {'abc': [1]},
        ...                            d_weak_order_share={'a~b>c': Fraction(2, 7), 'c>a~b': Fraction(4, 7)})
        >>> strategy = StrategyOrdinal({'abc': 'a'}, profile=profile)
        >>> print(strategy.tau)
        <a: 1/7, ab: 2/7, c: 4/7> ==> c
    """
    pass


def test_welfare():
    """
        >>> profile = ProfileHistogram({'a~b>c': Fraction(1, 3), 'a>b~c': Fraction(2, 3)})
        >>> profile.d_candidate_welfare
        {'a': Fraction(1, 1), 'b': Fraction(1, 3), 'c': 0}
    """
    pass
